#|  
   module      : application
   owner       : Franz Weber
   author      : Franz Weber/Olaf Mueller
   explanation : 

	representations of expressions of the simple typed lambda
	calculus.

	expressions are a variant of 'simple-tree'. The nodes must be
	of type 'expression-node'. 

	this module uses environment dependent evalution by special
	variables.


    	Applications
        ---------------

        (defflavor application (functor argument) (expression-node)
            (:gettable-instance-variables functor argument)
            (:init-keywords :functor :argument)
        )

        applications have the positions 'functor and 'argument

                a. generic implementations
                --------------------------

                        compare
                        copy
                        deallocate
                        index-of, set-elem, get-elem,
                            transfer-elem-out, transfer-elem-in,
                            iterate-non-recursive-method
                        apply-sub
                        expression-type

                        beta-reduce
                        beta-eta-reduce
                        delta-reduce

   history     :

        20.04.89
        90.06.12 Franz Weber
                 flavor base version
        91.03.04 Franz Weber
                 switch to typed lambda calculus
        91.04.13 Franz Weber
                 beta reduction inserted
	92.06.04 Olaf Mueller
		 usable for unification

   errors      : equal between expressions does not work
   bugs        : encloses compare incorrect
                 environment handling ??
                 how is *exp-identity* handled
   todo        : the mapping identity may be implemented by
                 a key stack mapping with structure shareing
                 on copy

                 copying of expression nodes is not done

                 it is not clear, what compare returns.
                    (alpha conversion e.g. is considered)


                 the usage of types is inconsistent

                 :expression-type returns a type node
                 :type returns a type


|# 

(mod-require "data/glo/v2/simple-lin")
(mod-require "data/glo/v2/cop")
(mod-require "data/glo/v2/pointer")
(mod-require "data/tre/v2/simple-tre")
(mod-require "data/stack/v2/key-stack-list")
(mod-require "logic/v2/expression/declaration/impls/declaration")
(mod-require "logic/v2/expression/type/impls/type")
(mod-require "logic/v2/expression/basic-type/impls/basic-type")
(mod-require "logic/v2/expression/function-type/impls/function-type")
(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")

(mod-require "data/lis/v2/lis-list-readonly")


; -------------------- NEW
(defflavor application (functor argument) (expression-node)
  	(:gettable-instance-variables functor argument)
	(:settable-instance-variables functor argument)
        (:init-keywords :functor :argument)
)


; --------definit
; until now only immediate setting of binder, head, 
; arguments out of functor and arguments is implemented. 
; For reverse purposes do:
;   (setq functor (assemble-exp-head head [arguments :butlast]))
;   (setq argument [arguments :get-elem [arguments :last]])
(definitmethod (application) ()
 (if  [[functor :binder] :is-empty]
   (progn
      (setq binder (make-instance 'lis-list-readonly))
      (setq head [functor :head])
      (setq arguments [[functor :arguments] :snoc argument])
   )
   (progn
        ; this is the case of exp's that are not in beta-normal-form.
	(setq binder (make-instance 'lis-list-readonly))
	(setq head functor)
        (setq arguments [(make-instance 'lis-list-readonly) :snoc argument])
   )
)
)

(defmethod (application :compare) (right)
    (if (eql self right) 'equal
    (if (typep right 'application)
        (let ((res [functor :compare [right :functor]]))
            (if (equal res 'equal)
                [argument :compare [right :argument]]
                res
            )
        )
        (flavor-compare 'application (type-of right))
    )
    )
)

(defmethod (application :equal) (right)
    (or (eql self right)
        (and (typep right 'application)
             [:functor |:equal [right :functor]]
             [:argument |:equal [right :argument]]
       )
    )
)

#|
; ----- copies

(defmethod (application :copy-from) (from)
    (setq functor [[from :functor] :copy])
    (setq argument [[from :argument] :copy])
    self
)

(defmethod (application :deallocate) () 
    (setq functor nil)
    (setq argument nil)
)
|#

; ----- expression-nodes
(defmethod (application :apply-sub) (sub)
    (let ((f [:functor |:apply-sub sub])
          (a [:argument |:apply-sub sub]))
    (if (and (eql f [:functor]) (eql a [:argument]))
        self
        (make-instance (type-of self)
             :functor f
             :argument a
        )
    )
    )
)

(defmethod (application :expression-type) ()
    (let ((ft [:functor |:expression-type]))
        (when ft
            (let ((at [:argument |:expression-type]))
                (when (and at 
                           (typep ft 'function-type)
                           [ft :domain |:equal at]
                      )
                    [ft :range]
                )
            )
        )
    )
)

; ----- simple aggregate

; positions are 'functor and 'argument

(defmethod (application :index-of) (pos) pos)


(defmethod (application :get-elem) (i) 
    (case i
        (functor functor)
        (argument argument)
    )
)

#|
(defmethod (application :set-elem) (i e) 
    (case i
        (functor (setq functor e))
        (argument (setq argument e))
    )
)
|#

(defmethod (application :transfer-elem-out) (i) 
    (case i
        (functor (prog1 functor (setq functor nil)))
        (argument (prog1 argument(setq argument nil)))
    )
)

#|
(defmethod (application :transfer-elem-in) (i e) 
    (case i
        (functor (setq functor e))
        (argument (setq argument e))
    )
)
|#

(defiterationmethod (application :iterate-non-recursive-method) (f) 
    (funcall f functor 'functor)
    (funcall f argument 'argument))


; ----- reductions

(defmethod (application :beta-reduce) ()
     (let ((f [[:functor] :beta-reduce]))
     (if (typep f 'abstraction)
         ; beta redex
         (progn
             [*exp-env* :push-key f [:argument]]
             (prog1
                 [[f :expression-node] :beta-reduce]
                 [*exp-env* :pop]
             )
         )
         ; simple copy
         (make-instance (type-of self) 
             :functor f
             :argument [[:argument] :beta-reduce])
     )
     )
)

(defmethod (application :beta-eta-reduce) ()
     (let ((f [[:functor] :beta-eta-reduce]))
     (if (typep f 'abstraction)
         ; beta redex
         (progn
             [*exp-env* :push-key f [:argument]]
             (prog1
                 [[f :expression-node] :beta-eta-reduce]
                 [*exp-env* :pop]
             )
         )
         ; simple copy
         (make-instance (type-of self) 
             :functor f
             :argument [[:argument] :beta-eta-reduce])
     )
     )
)

; an experiment (but doesn't work)
(defmethod (application :delta-reduce) ()
     (if (typep [:functor] 'abstraction)
         ; beta redex
         (progn
             [*exp-env* :push-key [:functor] [:argument]]
             (prog1
                 [[[:functor] :expression-node] :delta-reduce]
                 [*exp-env* :pop]
             )
         )
         ; simple copy
         [[[:functor] :delta-reduce] :eval [:argument] (type-of self)]
     )
)

