#|
   module      : expression-node
   owner       : Franz Weber
   author      : Franz Weber/Olaf Mueller
   explanation : 

        representations of expressions of the simple typed lambda
        calculus, adapted for unification.

        expressions are a variant of 'simple-tree'. The nodes must be
        of type 'expression-node'. 

        this module uses environment dependent evalution by special
        variables.
	
	to do: knew name for expression: expression-node




        Expression specific Methods
        -------------------------------


                (defflavor expression (binder-i head-i arguments-i) 
                         (parse-gen aggregate))
			(:init-keywords :binder :head :arguments)


                (defmethod (expression :encloses-compare) (right)

                returns 

                        'less           if right encloses self
                        'greater        if self encloses right
                        'uncompareable  else

                (defmethod (expression :encloses) (right))

                returns true, if self encloses right

                (virtualmethod (expression :apply-sub) (sub)

                returns a recursive copy (important!) of the expression node on
                which the substitution 'sub' was applied. Note, that
                it is not possible to do substitutions for expression
                modes in situ (this is different for whole expressions).
                Consider a variable. It must be replaced by it's
                substituent.

                (virtual-method (expression :expression-type) () ... )

                compute the effective type of the expression as a type
                node. If the expression isn't well typed, it returns
                nil.

                ----- reductions

                The copy and share rules are equivalent to copy for
                reductions. Reduced expression nodes don't (visibly)
                share with their sources. 

                (virtualmethod (expression :beta-reduce) () ...)

                make a beta reduction on self after copying..

                (virtualmethod (expression :beta-eta-reduce) () ...)

                make a beta-eta reduction on self after copying.

                (virtualmethod (expression :delta-reduce) () ...)

                make a beta, eta and delta  reduction on selfafter copying. 
                this is for future expansion, not implemented yet.
                this allows the invocation of interpreters.

                for future expansion, not implemented yet

                (defmethod (expression :eval) (arg standard-result-type) ... )

                hook for implementing delta reductions
                this may be redefined for special references, which do
                actual computing.

                for future expansion, not implemented yet


                ; ----- debugging

                inherited from parse-gen


		; --------------  N  E  W   ----------------

		-------- type
		(defmethod (expression :expression-type) () )
		
		 ------- concerning unification
		(defmethod (expression :is-rigid) () [head :is-constant])
		(defmethod (expression :is-flexible) () (not [:is-rigid]))

		(defmethod (expression :toplevel-eta-expand) () )

		(defmethod (expression :alpha-test) (r) )


		----------------------------------------------
			constructor functions
		----------------------------------------------

		; combines head and list of arguments
		(defun assemble-exp-head (head arguments))

		; combines binder as a list of abstractions and a body (expression-node)
		(defun assemble-exp-binder (abstractions body))

		; combination of assemble-exp-head and assemble-exp-binder
		(defun assemble-exp (abstractions head arguments))

   history     :
  
        20.04.89
	90.06.12 Franz Weber
		 flavor base version
        91.03.04 Franz Weber
		 switch to typed lambda calculus
	91.04.13 Franz Weber
		 beta reduction inserted
	92.06.04 Olaf Mueller
		 usable for unification

   errors      : equal between expressions does not work
   bugs        : encloses compare incorrect
		 environment handling ??
		 how is *exp-identity* handled
   todo        : the mapping identity may be implemented by
		 a key stack mapping with structure shareing
		 on copy

		 copying of expression nodes is not done

		 it is not clear, what compare returns.
		    (alpha conversion e.g. is considered)


                 the usage of types is inconsistent

		 :expression-type returns a type node
		 :type returns a type




|# 

(mod-require "data/glo/v2/simple-lin")
(mod-require "data/glo/v2/cop")
(mod-require "data/glo/v2/pointer")
(mod-require "data/tre/v2/simple-tre")
(mod-require "data/stack/v2/key-stack-list")
(mod-require "logic/v2/expression/declaration/impls/declaration")
(mod-require "logic/v2/expression/type/impls/type")
(mod-require "logic/v2/expression/basic-type/impls/basic-type")
(mod-require "logic/v2/expression/function-type/impls/function-type")
(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")

(mod-require "logic/v2/substitution/substitution-hash")

; ---------------- New  --------
(defflavor expression-node (binder head arguments) 
			   (parse-gen simple-linear-aggregate copies)
	(:gettable-instance-variables binder head arguments) 
	(:init-keywords :binder :head :arguments)
)

; ------------------- the environment

(defvar *exp-identity*)
(defvar *exp-env*)
  (setq *exp-identity* (make-instance 'key-stack-list))
  (setq *exp-env* (make-instance 'key-stack-list))

(proclaim '(special *exp-identity* *exp-env*))


; ----- own methods
(defmethod (expression-node :encloses-compare) (right)
   (if (eql right self)
       'equal
       (if [:find right :recursive t :test 'eql]
           'greater
           (if [right :find self :recursive t :test 'eql]
               'less
               'uncompareable
           )
       )
   )
)

(defmethod (expression-node :encloses) (right)
    (equal 'greater [:encloses-compare right])
)

; ----- substitution
(virtualmethod (expression-node :apply-sub) (sub))

; ----- reduction

(virtualmethod (expression-node :beta-reduce) ())
(virtualmethod (expression-node :beta-eta-reduce) ())
(virtualmethod (expression-node :delta-reduce) ())
(defmethod (expression-node :eval) (arg standard-result-type)
    (make-instance standard-result-type
        :functor self
        :argument [arg :delta-reduce]
    )
)



; ------------------------------------------
; --------------  N  E  W   ----------------
; ------------------------------------------

; -------- type
(virtualmethod (expression-node :expression-type) () )

;copy not allowed
(defmethod (expression-node :copy) () self)
(defmethod (expression-node :copy-from) () self)


; ----------------------------------------------
;	constructor functions
;----------------------------------------------

; combines head and list of arguments
(defun assemble-exp-head (head arguments) 
 	[arguments :reduce (x) (p head)
                (make-instance 'application 
			       :functor p 
			       :argument x)])
  

; combines binder as a list of abstractions and a body (expression-node)
(defun assemble-exp-binder (abstractions body)
    (let ((new-abss nil)
	  (sub-exp (make-instance 'substitution-hash))
         )
    ; make a copy  of the arguments
    ; note that the expression nodes of the abstraction are not copied 

    
    (setq new-abss 

	  [abstractions :map (x) 
	      (let ((res (make-instance 'abstraction 
					:type        [x :type]
					:picture     [x :picture]
					:is-constant [x :is-constant]))
					)
		  ; @@@ or vice versa ?
		  ;[*exp-identity* :push-key x res]
                   [sub-exp :insert-key x 
                        (make-instance 'reference :ref res)]
                  res)])
    (setq body [body :apply-sub sub-exp])
    [new-abss :iterate (x) 
	;[*exp-identity* :pop]
	[x :set-expression-node body] (setq body x)
    ]
    ; now body contains the desired expression
    body))



; combination of assemble-exp-head and assemble-exp-binder
(defun assemble-exp (abstractions head arguments)
	(assemble-exp-binder abstractions (assemble-exp-head head arguments)))
 

; ----------------------------------------------------------------------
;                     unification utilities
; ----------------------------------------------------------------------

; -------- head checking
(defmethod (expression-node :is-rigid) () [head :is-constant])
(defmethod (expression-node :is-flexible) () (not [:is-rigid]))


; make shure that selfs head has as many arguments as indicated by its type 
(defmethod (expression-node :toplevel-eta-expand) () 
 (let ((arg-types [[head :expression-type] :argument-types])
       (idx nil)
       (bin binder)
       (arg arguments)
      )
     ; first throw away the elements that are already represented 
     ; by a corresponding argument
     (setq idx [arg-types :first])
     (dotimes (x [arguments :count]) (setq idx [arg-types :next idx]))
     ; now fill up arguments and abstractions
     (when idx
     [arg-types :iterate :from `(at-index ,idx) (x)
         (let ((a (make-instance 'abstraction :type x 
	 			              :picture (uni-new-pic "x"))))
	   (setq bin [bin :cons a])
	   (setq arg [arg :snoc (make-instance 'reference :ref a)])
         )
     ]
     )
     (assemble-exp bin head arg)
 )
)


; Test if binder and head are equal modulo alpha conversion
; Heads of beta normalized expressions are always references
; They are alpha convertible if they refer to the same declarations
; or they have the same position in the binder @tested
(defmethod (expression-node :alpha-test) (r) 
  (let ((h1 [  :head])
	(h2 [r :head]))
    (or [h1 :equal h2] ; h1, h2 are ref's to the same external variable
	(let* ((a1 [  :binder])
	       (a2 [r :binder])
	       (i1 [a1 :find [h1 :ref]]) 
	       (i2 [a2 :find [h2 :ref]]))
	  (and i1 ; h1 is locally bound
	       i2 ; h2 is locally bound
	       (= [a1 :index-pos i1] 
		  [a2 :index-pos i2]))))))



