
#|)
   module      : parse
   owner       : Franz Weber
   author      : Franz Weber
   explanation : parsing and unparsing of expressions

	1. Lexical Conventions
	----------------------

		- Comments are enclosed by curly braces '{' '}'
		  They may be nested

                - if a line contains a ';' , the rest of the line is a
		  comment

		- non empty sequences of comments, newlines, tabs and spaces
		  are refered to as 'white space'

		- There are the following tokens:

			single character:

			( ) . : eof

			identifiers: sequences of characters not
			containing white space or a single character token

               - Tokens may but need not to be seperated by white space
		 except they are both identifiers

	2. Grammar
	----------

		expression ::= application | abstraction | symbol

		application ::= '(' exp exp ')'

		abstraction ::= declaration ':' exp

		declaration ::= symbol '.' type

		symbol ::= identifier


		type ::= basic-type | function-type

		basic-type ::= identifier

		function-type ::= '(' type  type ')'

		    ; the left type is the range, the right type the domain


		environment ::= '(' (variable-declaration 
				    | constant-declaration) * ')'

                variable-declaration ::= declaration

		constant-declaration ::= declaration ":"


		substitution ::= '(' (symbol expression)* ')'

	       

    	3. Semantic
	-----------

		- types are globally identified, when they have the
		  same string. 

                - the function type (R D) means the type with domain
		  type D and range type R !!!!!!!!!

	4. Examples
	-----------

	expressions :

	x.B: x	identity on B

	x.(B C): y.C: (x y)

	types :

	(A B)           functions from B to A

	environments :

	(x.(A B) y.B:)

	substutions

	(x (f y) y (r s)) the substitution, which replaces x by (f y)
	                  and y by (r s)

			  note, that only variables, which are
			  declared in an environment, make sense.
			  (You may insert abstractions in
			  environments!)


	5. Interface
	------------


	(defun rul-for-parse-expression 
	    (item 
	    &key 
	     from-file 
	     (map (make-instance 'key-stack))
	     (env *exp-environment*))
	    )

	map may contain declarations of symbols which are used by the
	parser. It must be a key-stack. env may be an environment,
	which is automatically pushed on the map.


	(defun rul-for-parse-type
	    (item &key from-file))

	(defun rul-for-parse-environment
	    (item &key from-file))

	(defun rul-for-parse-substitution
	    (item 
	    &key 
	    from-file
	    (map (make-instance 'key-stack))
	    (env *exp-environment*)
	    ))

        For all parse functions:
	
	if from-file is nil, item must be a string. The string gets
	parsed and the resulting expression is returned. If from-file
	is not nil, item is interpreted as a file name and the file
	gets parsed.

	6. Unparsing
	------------

	For expressions and expression node and type and type node:

	(defmethod (<...> :unparse) (&optional (stream t)) ... )

	unparsing obeys the rule, that variables, which get in the
	scope of a inner variable with the same name, one of them
	must be renamed. For a variable with name "x" this is done by
	trying "x'" "x''" "x*" "x1 "x2" "x3" ... Always the inner
	bound variable  is renamed (of course)

   history     :
  
        20.04.89
	90.06.12 Franz Weber
		 flavor base version
        91.03.19 Franz Weber
		 first typed lambda calculus version
	91.04.17 Franz Weber
		 unparse uniques variable names, if necessary
	91.04.18 Franz Weber
		 parsing of types and environments
		 type nodes instead of types
	91.05.16 parsing of substitutions
        92.06.04 no parsing of expressions or types, only nodes

   errors      :
   bugs        : 
   todo        :
|# 

(mod-require "data/glo/v2/simple-lin")
(mod-require "data/glo/v2/cop")
(mod-require "logic/v2/expression/environment/impls/environment")
(mod-require "logic/v2/substitution/substitution-hash")


(defmacro if-unparse (object stream)

   `(if (and (instancep ,object)
	     [,object :operation-handled-p :unparse]
	)
	[,object :unparse :stream ,stream]
	(format ,stream "#~a#" ,object)
    )
)

(defun rul-for-init ())

; comments are enclosed in '{}', they may be nested
(defun rul-for-skip-comment (stream)
    (loop
        (read-char stream)
	(case (peek-char t stream nil 'eof)
	    (#\} (read-char stream) (return-from rul-for-skip-comment))
	    (#\{ (rul-skip-comment stream))
	    ('eof (error "for parse: eof occured in comment"))
	)
    )
)

(defun rul-for-next-token (stream)
   (loop ; skip white space
      (case (peek-char t stream nil 'eof) 
	 (#\; (read-line stream))
	 (#\{ (rul-for-skip-comment stream))
	 (t (return))
      )
   )
   (case (peek-char nil stream nil 'eof)
       (#\. (read-char stream) (return-from rul-for-next-token '(dot)))
       (#\: (read-char stream) (return-from rul-for-next-token '(colon)))
       (#\( (read-char stream) (return-from rul-for-next-token '(left-paranthesis)))
       (#\) (read-char stream) (return-from rul-for-next-token '(right-paranthesis)))
       ('eof (return-from rul-for-next-token '(eof)))
       (t 
	   (list
	       'identifier
	       (let ((s ""))
		   (loop
		       (setq s (cat s (string (read-char stream))))
		       (when (member (peek-char nil stream nil 'eof) 
				     '(eof #\Space #\Tab #\: #\( #\) #\; 
				       #\.  #\{ #\} #\Newline))
			   (return)
		       )
		   )
		   s
	       )
	   )
       )
   )
)

(defvar *rul-for-type-table* (make-hash-table))

; ----- kernel parsing routine
(defun rul-for-parse-kern 

    (kind 	; may be type, expression or environment (as symbols)
     item 	; string or filename
    &key 
     (from-file nil) 	; indicates, wether string is a filename
     (map (make-instance 'key-stack-list))	; predeclared identifiers
     (env (make-instance 'environment-lis))
    &aux 
     tok	; look ahead
    )

(labels

    ((parse (stream)
    (labels
    ((next () (setq tok (rul-for-next-token stream)))
     (parse-type ()
	  (case (car tok)
	     ('left-paranthesis
		 (next) ; skip left paren
		 (prog1
		    (make-instance 'function-type
			:range (parse-type)
			:domain (parse-type)
		    )
		    (next) ; skip right paranthesis
		 )
	     )
	     ('identifier
		 (prog1
		    (make-instance 'basic-type :picture (cadr tok))
		    (next)
		 )
	     )
	     (t (error "illegal type"))
	 )
      )
      (parse-expression ()
	 (case (car tok)
	     ('identifier (let ((id (cadr tok)))
			      (next)
			      (if (equal (car tok) 'dot)
				  (let ((res (make-instance 'abstraction 
							    :picture id)))
				     (next) ; skip '.'
				     [res :set-type (parse-type)]
				     (next) ; skip ':'
				     [map :push-key id res]
				     [res :set-expression-node (parse-expression)]
				     [map :pop]
				     res
				  )
				  (if [map :get id :test 'equal]
				      (make-instance 'reference :ref 
				          [map :get id :test 'equal]
				      )
				      (throw 
					  'parse-error 
					  (cat "undeclared identifier " id)
				      )
				  )
			      )
			  )
	     )
	     ('left-paranthesis
		 (next)
		 (prog1
		    (make-instance 'application
			:functor (parse-expression)
			:argument (parse-expression)
		    )
		    (next) ; skip right paranthesis
		 )
	     )
	     ('eof (error "for parse: end of file occured in formula-node"))
	     (t (error "unkown token ~A~%" tok))
	 )
     )
     (parse-environment ()
        (case (car tok)
	     ('left-paranthesis
		 (next) ; skip left paren
		 (let ((res (make-instance 'environment-lis)))
		    (loop
			(case (car tok)
			    ('right-paranthesis (return))
			    ('identifier
				(let* ((picture (prog1 (cadr tok) 
						       (next) ;skip id
						       (next) ; skip point
						       ))
				       (type (parse-type))
				       (is-constant
					   (when (equal 'colon (car tok))
					       (next)
					       t
					   )
				       )
				      )
				  [res :declare 
				      type
				      picture 
				      is-constant]
				)
			    )
			    (t
(error 
"illegal environemnt: ~a intstead of identifier or right paraanthesis~%" 
    (cadr tok)
)
			    )
			)
		    )
		    (next) ; skip right paranthesis
		    res
		 )
	     )
	     (t (error "illegal environment"))
	 )
      )
      (parse-substitution ()
        (case (car tok)
	     ('left-paranthesis
		 (next) ; skip left paren
		 (let ((res (make-instance 'substitution-hash)))
		    (loop
			(case (car tok)
			    ('right-paranthesis (return))
			    ('identifier
				(let ((dec [map :get (cadr tok)]))
				    (if dec
				      (progn
				      (next) ; skip identifier
				      [res :move-in-key
					  dec
					  (parse-expression)
				      ]
				      )
				      (progn
				      (error
"undeclared identifier ~a~%" (cadr tok)
				      )
				      (next) ; skip identifier
				      )
				   )  
				)
			    )
			    (t
(error 
"illegal substitution: ~a intstead of identifier or right paraanthesis~%" 
    (cadr tok)
)
			    )
			)
		    )
		    (next) ; skip right paranthesis
		    res
		 )
	     )
	     (t (error "illegal substitution"))
	 )
      ) ; parse-sub
      ) ; labels head
	  (next)
	  (catch 'parse-error
	      (case kind
		  ((expression substitution) 
		      [env :iterate (x) [map :push-key [x :picture] x]]
		      (prog1
			 (if (equal kind 'substitution)
			     (parse-substitution)
			     (parse-expression)
			 )
			 [env :iterate (x) [map :pop]]
		      )
		  )
		  (type (parse-type))
		  (environment (parse-environment))
		  (t (error "unknown kind"))
	      )
	  )
      ) ; inner lables
      )
    ) ; outer label head

    (if from-file
	(with-open-file (s item) (parse s))
	(with-input-from-string (s item) (parse s))

    )
) ;outer labels
)

(defun rul-for-parse-expression 
    (item 
    &key 
     from-file 
     (map (make-instance 'key-stack-list))
     (env (if *exp-environment* *exp-environment* (make-instance 'environment-lis)))
    )

    (rul-for-parse-kern 'expression item :from-file from-file :map map :env env)
)

(defun rul-for-parse-substitution 
    (item 
    &key 
     from-file 
     (map (make-instance 'key-stack-list))
     (env (if *exp-environment* *exp-environment* (make-instance 'environment-lis)))
    )

    (rul-for-parse-kern 'substitution item :from-file from-file :map map :env env)
)

(defun rul-for-parse-type 
    (item &key from-file)

    (rul-for-parse-kern 'type item :from-file from-file :map nil)
)

(defun rul-for-parse-environment
    (item &key from-file)

    (rul-for-parse-kern 'environment item :from-file from-file :map nil)
)

; -------------------- Unparsing
(defvar *exp-names* (make-instance 'key-stack-list))
(proclaim '(special *exp-names*))

; expressions

#|
(defmethod (expression :unparse) 
    (&key (stream t) 
	  (env *exp-environment*))
    (when env [env :iterate (x) [*exp-names* :push-key x [x :picture]]])
    (if-unparse [:content] stream)
    (when env [env :iterate (x) [*exp-names* :pop]])
)
|#

; -------------------- formula-nodes

(virtualmethod (expression-node :unparse) (&key (stream t)))

; -------------------- references

(defmethod (reference :unparse) (&key (stream t))
    (let ((name [*exp-names* :get ref :test 'eql]))
	(unless name 
	    (setq name (rul-for-parse-search-name 
		           *exp-names* 
			   (cat "#" [ref :picture] "#")))
	)
	(format stream "~a" name)
    )
)

; -------------------- abstraction


; conventions for picture renaming
(defun rul-for-parse-search-name (map str)
    (labels
     ((unused (nam)
	 (not [map :exists ((key val)) (equal val nam)])
      )
     )

       (or
	   (when (unused str) str)
	   (when (unused (cat str "'")) (cat str "'"))
	   (when (unused (cat str "''")) (cat str "''"))
	   (let ((i 0))
	      (loop
		  (let ((n (cat str (format nil "~a" i))))
		     (when (unused n) (return n))
		     (incf i)
		  )
	      )
	  )
       )
    )
)

(defmethod (abstraction :unparse) (&key (stream t))
    (if (stringp [:picture])
       (progn
       [*exp-names* :push-key
	   self 
	   (rul-for-parse-search-name *exp-names* [:picture])]
       (format stream "~a" [*exp-names* :get self :test 'eql])
       )
       (format stream "#~a#" [:picture])
    )
    (format stream ".")
    (if-unparse [:type] stream)
    (format stream ": ")
    (if-unparse [:expression-node] stream)
    [*exp-names* :pop]
)
; -------------------- application

(defmethod (application :unparse) (&key (stream t))
    (format stream "(")
    (if-unparse [:functor] stream)
    (format stream " ")
    (if-unparse [:argument] stream)
    (format stream ")")
)

; -------------------- basic types

(defmethod (basic-type :unparse) (&key (stream t))
    (format stream "~a" [:picture])
)

; -------------------- function types

(defmethod (function-type :unparse) (&key (stream t))
    (format stream "(")
    (if-unparse [:range] stream)
    (format stream " ")
    (if-unparse [:domain] stream)
    (format stream ")")
)

; -------------------- types

;(flavor-mixin type content :unparse)

