#|
   module      : reference
   owner       : Franz Weber
   author      : Franz Weber/Olaf Mueller
   explanation : 

	representations of expressions of the simple typed lambda
	calculus.

	expressions are a variant of 'simple-tree'. The nodes must be
	of type 'expression-node'. 

	this module uses environment dependent evalution by special
	variables.


        References
        -------------

        (defflavor reference () (expression-node))

                note, that the content is always a declaration
                this version allows only abstractions as declarations

                a. generic implementations
                --------------------------

                        compare
                            does not work with declarations
                        copy-from
                        apply-sub
                        expression-type

                        beta-reduce
                        beta-eta-reduce
                        delta-reduce

                        attribute-string

                Note, that the implementations for the methods of 
                simple-aggregate are inherited by pointer.

                b. reference specific methods
                -----------------------------

                        is-free-in
                        is-constant
                        is-variable
                        ref              reads the content

   history     :
  
        20.04.89
	90.06.12 Franz Weber
		 flavor base version
        91.03.04 Franz Weber
		 switch to typed lambda calculus
	91.04.13 Franz Weber
		 beta reduction inserted
        92.06.04 Olaf Mueller
                 extended for unification

   errors      : equal between expressions does not work
   bugs        : encloses compare incorrect
		 environment handling ??
		 how is *exp-identity* handled
   todo        : the mapping identity may be implemented by
		 a key stack mapping with structure shareing
		 on copy

		 copying of expression nodes is not done

		 it is not clear, what compare returns.
		    (alpha conversion e.g. is considered)


                 the usage of types is inconsistent

		 :expression-type returns a type node
		 :type returns a type


|# 

(mod-require "data/glo/v2/simple-lin")
(mod-require "data/glo/v2/cop")
(mod-require "data/glo/v2/pointer")
(mod-require "data/tre/v2/simple-tre")
(mod-require "data/stack/v2/key-stack-list")
(mod-require "logic/v2/expression/declaration/impls/declaration")
(mod-require "logic/v2/expression/type/impls/type")
(mod-require "logic/v2/expression/basic-type/impls/basic-type")
(mod-require "logic/v2/expression/function-type/impls/function-type")
(mod-require "logic/v2/expression/parse-gen/impls/parse-gen")

(mod-require "data/lis/v2/lis-list")
(mod-require "data/lis/v2/lis-list-readonly")


; -------------------- references

(defflavor reference () (pointer expression-node))



; ------ definit 
(definitmethod (reference) ()
	(setq binder (make-instance 'lis-list-readonly))
	(setq head self)
	(setq arguments (make-instance 'lis-list-readonly))
)


; ----- comparison
(defmethod (reference :compare) (right)
    (if (eql self right) 'equal
    (if (typep right 'reference)
         (if (and (typep ref 'abstraction)
                  (typep [right :ref] 'abstraction))
        (let ((id   (or [*exp-identity* :get ref :test 'eql] 
                        ref))
              (rid  (or [*exp-identity* :get [right :ref] :test 'eql] 
                        [right :ref]))
             )
             (if (eql id rid)
                 'equal
                 [id :encloses-compare rid]
             )
         )
         (if (typep ref 'abstraction)
             'greater
             (if (typep [right :ref] 'abstraction)
                 'less
                 [ref :compare [right :ref]]
             )
         )
        )
        (flavor-compare 'reference (type-of right))
    )
    )
)

(defmethod (reference :equal) (right)
    (or (eql self right)
        (and
            (typep right 'reference)
            (let ((s (or [*exp-identity* :get [:ref] :test 'eql] [:ref]))
                  (r (or [*exp-identity* :get [right :ref] :test 'eql] 
                         [right :ref]))
                 )
                (eql s r)
             )
        )
    )
)

#| old
(defmethod (reference :equal) (right)
    (or (eql self right)
        (and
            (typep right 'reference)
            [:ref |:equal [right :ref]
        )
    )
)

|#
; ----- copies
; @@@@ false @@@
;(defmethod (reference :copy-from) (from)
    ;(setq ref [*exp-identity* :get [from :ref] :test 'eql])
    ;(unless ref (setq ref [from :ref]))
    ;self
;)


; ----- simple aggregate

; inherited from pointer
(defmethod (reference :attribute-string) () [:ref |:picture])

; ----- expression-nodes
(defmethod (reference :apply-sub) (sub)
    (let ((f [sub :get ref :test 'eql]))
        (if f
            [f :apply-sub sub]
            (let ((r [*exp-identity* :get ref :test 'eql]))
		(if r
		    (make-instance 'reference :ref r)
		    self
		)
            )
        )
    )
)

(defmethod (reference :expression-type) () [:ref | :type])

; ----- reductions
(defmethod (reference :beta-reduce) () 
    (let ((sub [*exp-env* :get [:ref] :test 'eql]))
       (if sub
           [sub :beta-reduce]
           [ :copy]
       )
    )
)

(defmethod (reference :beta-eta-reduce) () 
    (let ((sub [*exp-env* :get [:ref] :test 'eql]))
       (if sub
           [sub :beta-eta-reduce]
           [ :copy]
       )
    )
)

(defmethod (reference :delta-reduce) () 
    (let ((sub [*exp-env* :get [:ref] :test 'eql]))
       (if sub
           [sub :delta-reduce]
           [:copy]
       )
    )
)

; ----- reference specific methods
(defmethod (reference :is-free-in) (exp)
    (not [exp :encloses ref])
)

(defmethod (reference :is-constant) () (not [:is-variable]))

(defmethod (reference :is-variable) () [[:ref] :is-variable])

(defmethod (reference :copy) ()
    (setq r [*exp-identity* :get ref :test 'eql])
    (if r 
        (make-instance (type-of self) :ref r)
        self
    )
)


