#|
   module      : aggregate
   owner       : Franz Weber
   author      : Franz Weber
   explanation : general functions of abstract aggregate like data types
		 See the book of Uhl/Endres
                 See the document
                     'A Catalogue of Abstract Data Types in Flavors'
                 See the document
                     'Aggregates'

	Aggregates add to simple aggregates three concepts:

	1. A concept of construction and destruction of aggregates

	2. A concept of transforming s-exp to aggregates

	3. Two special Iterators based on the destructors.


	1. Constructors
	---------------

	Constructors and destructors are using 'positions' to identify
	the indices where elements (component objects) should be
	inserted or deleted. Positions are defined by descendent
	classes. Positions server to denote 'indices', which may even
	not exist or are very abstract (insert an element at the
	'first position of a list', the index is generated by this
	action, therefore there is no name for it before this
	action).

	There are four basic kind of constructions and destrcuctions:

		1. insert 
		       don't destroy with source
		       don't clear target before action

		1. move 
		       destroy source
		       don't clear target before action

		1. transfer 
		       destroy source
		       clear target before action

		1. copy 
		       don't destroy with source
		       clear target before action

	It is not specified for these operations, if the elements
	themselves are copied too, when you copy the aggregate
	comprising them. Both variants are possible. However, normally
	they are not copied (This is in contrast to the catalogue of
	J.Uhl und H.A. Schmidt).

	(virtualmethod (aggregate :clear) () (virtual))
	         set the aggregate to empty
	(defmethod (aggregate :deallocate ) () ...)
	         reset the aggregate to empty and free all
		memory, which is needed for the aggregate
		(object afterwards not usable)

	(defmethod (aggregate :insert) (e &key (to-pos 'dont-care)) ...)
		insert elem at position
	(defmethod (aggregate :insert-aggregate) (a))
		insert all component objects of aggregate a
	(defmethod (aggregate :find-or-insert) (e &key (to-pos 'dont-care) test)
		insert e if e isn't an element of the aggregate
		second value of result is nil, when object wasn't
		an element of the aggregates

	(defmethod (aggregate :remove) (i) ...)
		remove element at index i
	(defmethod (aggregate :remove-elem) (e &key test)
		remove (one, first) element which is equal to e
	(defmethod (aggregate :remove-all) (e &key test part from to)
		remove all elements, which are equal to e
	
	- move in or between objects
	(virtualmethod (aggregate :move) (from-pos &key to-pos from)...)
		move object from from-pos  to to-pos. The 'from'
		object  may be situated in object 'from' (otherwise
		self).
	(defmethod (aggregate :move-aggregate) (a &key to-pos)
		move the whole aggregate a to self

	- move in or out of variables
	(virtualmethod (aggregate :move-out) (pos) (virtual))
		move out (and return) object at pos
	(virtualmethod (aggregate :move-in) (e &key to-pos) (virtual))
		move in e at to-pos

	(defmethod (aggregate :transfer) ()
		return a copy of self and clear self
	(defmethod (aggregate :transfer-from) (from)
		make self identical to from and clear from

	(defmethod (aggregate :flatten) ()
		flatten self
	(defmethod (aggregate :flatten-from) (from &key (copy nil)))
		make self a copy of the flattend version of from.
		from may be copied or cleared by this method.
		'flatten' means to insert all component objects of
		the component objects of from to self.
	

	2. S-Expressions
	----------------

	The descendent classes make rules, how to code their
	aggregates by s-exp (writeable lisp representations). 

	(virtualmethod (aggregate :insert-s-exp) (s &key (to-pos nil to-pos-p)))
		insert elem coded as s-exp at position
	(virtualmethod (aggregate :move-in-s-exp) (e &key to-pos) (virtual))
		moves in the s-exp e
	(defmethod (aggregate :copy-from-s-exp) (exp) [:clear] [:insert-s-exp exp])
		clears self before insertiing the s-exp
	(defmethod (aggregate :transfer-s-exp) (exp)
		clears self before moving in the s-exp

		    
   history     : 90.06.26 Version 1
		 90.09.26 Version 2
   errors      :
   bugs        : parts are not considerd
		 'to-pos is not fully correct'
		 default implementations ofinsert-struct and move-struct are
		 not for all cases correct.
		 init methode sollte liste von elementen 
		 als argument haben koennen
		 insert multiple is virtual
		 remove-all should be an iteration-method
|#

(mod-require "flavor/v4")
(mod-require "data/glo/v2/cop")
(mod-require "data/glo/v2/simple-aggregate")

(defflavor aggregate () (simple-aggregate copies) (:init-keywords :init-value))

; -------------------- constructors (page 57)
(definitmethod (aggregate) (&key (init-value nil init-value-p)))

(virtualmethod (aggregate :clear) () (virtual))
(defmethod (aggregate :deallocate ) () [:clear])
    ; returns the object
(defmethod (aggregate :insert) (e &key (to-pos 'dont-care)) 
    [:move-in (if-copy e [e :copy])  :to-pos to-pos])
    ; returns the index
(virtualmethod (aggregate :insert-s-exp) (s &key (to-pos nil to-pos-p)))
; default implementation is faulty
(defmethod (aggregate :insert-struct) (s)
    [s :iterate (el) [:insert el]]
)
(defmethod (aggregate :find-or-insert) (e &key (to-pos 'dont-care) test)
    (let ((ff [:find e :test test]))
	(if ff
	    (values ff t)
	    (values [:insert e :to-pos to-pos] nil)
	)
    )
)
    ; returns index and 'found'
(defmethod (aggregate :remove) (i) [:move-out i])
    ; argument is remove-position
(defmethod (aggregate :remove-elem) (e &key test) 
    [:remove `(at-index ,[:find e :test test])])
    ; argument is an element
(defmethod (aggregate :remove-all) (e &key test part from to)
     [:iterate 'move :part part :from from :to to
	 (el in)
	 (when (if test (funcall test el e) (flavor-equal el e)) 
	     [:remove `(at-index ,in)])
     ]
)
; generic remove is omitted
(virtualmethod (aggregate :move) (from-pos &key to-pos from) (virtual))
    ; to-pos and from are optional
(defmethod (aggregate :move-struct) (s &key to-pos)
    [s :iterate 'move (el in) [:move `(at-index ,in) :from s]]
)
; move to and from variables
(virtualmethod (aggregate :move-out) (pos) (virtual))
(virtualmethod (aggregate :move-in) (e &key to-pos) (virtual))
(virtualmethod (aggregate :move-in-s-exp) (e &key to-pos) (virtual))
(defmethod (aggregate :copy-from-s-exp) (exp) [:clear] [:insert-s-exp exp])
(defmethod (aggregate :transfer) () 
    [(make-instance (type-of self)) :transfer-from self]
)
(defmethod (aggregate :transfer-from) (from) 
    [:copy-from from]
    [from :clear]
    self
)
(defmethod (aggregate :transfer-s-exp) (exp) [:clear] [:move-in-s-exp exp])

(defmethod (aggregate :flatten) ()
    (let ((new [(make-instance (type-of self)) :flatten-from self]))
	[:transfer-from new]
	[new :deallocate]
    )
    self
)

(defmethod (aggregate :flatten-from) (from &key (copy nil))
    [:clear]
    (if copy
	[from :iterate (el) [:insert-struct el]]
	[from :iterate 'move (el) [:move-struct el]]
    )
    self
)

; -------- special iterations

(defiterationmethod (simple-aggregate :remove-min) ()
    [:remove `(at-index ,(multiple-value-bind (el in) 
        [:calliterationmethod :min] in))]
)

(defiterationmethod (simple-aggregate :remove-max) ()
    [:remove `(at-index ,(multiple-value-bind (el in) 
        [:calliterationmethod :max] in))]
)
