#|
   module      : key-generic-aggregate
   owner       : Franz Weber
   author      : Franz Weber
   explanation : key aggregates as 'add on' on general aggregates

	Key aggreggates don't assume structure between component
	objects but within the component objects. They assume, that
	each component object consists of a pair '(<key> <info>) and
	that each key occurs only once in the whole aggregate
	(generalized mapping). Therefor 'key-generic-aggregate' is a
	orthogonal package, which may be added to all kinds of
	aggergates, which don't assume any structure for their
	component objects.

	The package defines a lot of operations also defined for
	normal aggregates but based on the identification of objects
	by their 'key-part' or their 'info-part'. Pairs itself are a
	simple abstract data type with operations

	(defun pair (key info) ...)	to create a pair
	(defun key-of (pair) ..)	to get the key part of a pair
	(defun info-of (pair) ...)	to get the info part of a pair

	This representation is only for external representations. Key
	aggregates may represent this pairs arbitrarily.

	1. Constructors
	---------------

	The following methods are the same as for aggregates,, except
	that a component object is given bye  key and info.

	(defmethod (key-generic-aggregate :insert-key) (k info)
	(defmethod (key-generic-aggregate :find-or-insert-key) 
	    (k info &key test)

	The following methods are the same as for aggregates,, except
	that the component object is identified by key or info.

	(defmethod (key-generic-aggregate :remove-key) (k &key test)
	(defmethod (key-generic-aggregate :remove-info) (i &key test)

	The following methods is the same as for aggregates,, except
	that a component object is given bye  key and info.

	(defmethod (key-generic-aggregate :move-in-key)
	    k i &key (to-pos nil to-pos-p))

	The following  method returns two values, the first is the key,
	the second the info of component object at position pos.

	(defmethod (key-generic-aggregate :move-out-key) (pos)


	2. Indices
	----------

	The following methods are the same as for aggregates,, except
	that key and info parts of component objects may be read or
	set seperatly.

	(defmethod (key-generic-aggregate :get-key) (i)
	(defmethod (key-generic-aggregate :get-info) (i)
	(defmethod (key-generic-aggregate :set-key) (i key)
	(defmethod (key-generic-aggregate :set-info) (i info)
	(defmethod (key-generic-aggregate :transfer-info-in) (i info)
	(defmethod (key-generic-aggregate :transfer-info-out) (i)
	(defmethod (key-generic-aggregate :modify-info) (i modification)

	The following methods are not available for key aggregates,
	because they may disturb consistency of key aggregates.

	(hidemethod (key-generic-aggregate :transfer-elem-in))
	(hidemethod (key-generic-aggregate :transfer-elem-out))
	(hidemethod (key-generic-aggregate :modify-elem))


	3. Selectors
	------------

        The following methods are the same as for aggregates,, except
	that the component objects are identified by key or info.

	(defiterationmethod (key-generic-aggregate :find-key) (k &key test)
	(defiterationmethod (key-generic-aggregate :find-info) (i &key test)
	(defiterationmethod (key-generic-aggregate :count-info) (i &key test)
	(defiterationmethod (key-generic-aggregate :count-key) (k &key test)
	(defiterationmethod (key-generic-aggregate :is-info) (i)
	(defiterationmethod (key-generic-aggregate :is-key) (k)

	(defiterationmethod (key-generic-aggregate :get) (k &key test)
	    The same as find key but additional selecting the info
	    part of the result.

	

	4. Iterators
	------------

	The following methods implement the key variant of iterators
	in 'aggregates':

	(defiterationmethod (key-generic-aggregate :iterate-key-method) (f)

	(defiterationmethod (key-generic-aggregate :reduce-key-method)
	   (operation &key (unit-key nil unit-p) (unit-info nil unit-info-p))

	A reduction only working on the key part of the key aggregate
	(defiterationmethod (key-generic-aggregate :reduce-key)
	    (operation &key (unit nil unit-p))

	A reduction only working on the info part of the key aggregate
	(defiterationmethod (key-generic-aggregate :reduce-info)
	    (operation &key (unit nil unit-p))
	     

	All iterators implemented in 'aggregates' are extended to key
	iterations.

	find-property, skip-property, count-property, exists, forall


	5. Order Dependent Operations
	-----------------------------

	The same as for aggregates on the info part of the key
	aggregate:

	(defiterationmethod (key-generic-aggregate :max-info) ()
	(defiterationmethod (key-generic-aggregate :min-info) ()


   history     : 90.08.07 Version 1
   errors      :
   bugs        :
|#

(mod-require "data/glo/v2/aggregate")
(defflavor key-generic-aggregate () (aggregate))

; -------------------- basic operations
(defun key-of (e) (car e))
(defun info-of (e) (cadr e))
(defun pair (key info) (list key info))

; -------------------- constructors (page 57)
(defmethod (key-generic-aggregate :insert-key) (k info) [:move-in-key k info])
    ; returns index
(defmethod (key-generic-aggregate :find-or-insert-key) (k info &key test)
    (let ((ff [:find-key k :test test]))
        (if ff
            (values ff t)
            (values [:insert-key k info] nil)
        )
    )    
)
    ; returns index and 'found'
(defmethod (key-generic-aggregate :remove-key) (k &key test) 
    [:remove [:find-key k :test test]])
(defmethod (key-generic-aggregate :remove-info) (i &key test) 
    [:remove [:find-info i :test test]])
(defmethod (key-generic-aggregate :move-in-key) 
    (k i &key (to-pos nil to-pos-p))
    (if to-pos-p
        [:move-in (pair k i) :to-pos to-pos]
        [:move-in (pair k i)]
    )
)
(defmethod (key-generic-aggregate :move-out-key) (pos)
    (values-list [:move-out pos])
)

; -------------------- Operations based on indices (page 73) 

;(hidemethod (key-generic-aggregate :set-elem))
(defmethod (key-generic-aggregate :get-key) (i) (key-of [:get-elem i]))
(defmethod (key-generic-aggregate :get-info) (i) (info-of [:get-elem i]))
(defmethod (key-generic-aggregate :set-key) (i key) 
    [:set-elem i (pair key [:get-info i])])
(defmethod (key-generic-aggregate :set-info) (i info) 
    [:set-elem i (pair [:get-key i] info)])
;(hidemethod (key-generic-aggregate :transfer-elem-in))
;(hidemethod (key-generic-aggregate :transfer-elem-out))
(defmethod (key-generic-aggregate :transfer-info-in) (i info) 
    [:set-elem i (pair [:get-key i] info)]
)
(defmethod (key-generic-aggregate :transfer-info-out) (i) 
    (prog1
	(info-of [:get-elem i])
	[:set-info i nil]
    )
)
;(hidemethod (key-generic-aggregate :modify-elem))
(defmethod (key-generic-aggregate :modify-info) (i modification)
    [:set-info i (funcall modification [:get-info i])]
)

; -------------------- Access by Position Count (page 76)

; -------------------- Selector Operations (page 77)
(defiterationmethod (key-generic-aggregate :find-key) (k &key test)
    [:calliterationmacro (item index) :reduce (res nil)
	(let ((key (key-of item)))
	   (if test
	      (when (funcall test key k) (return index))
	      (when (flavor-equal key k) (return index))
	   )
	)
    ]
)

(defiterationmethod (key-generic-aggregate :get) (k &key test) 
    (let ((res [:calliterationmethod :find-key k :test test]))
       (when res [:get-info res])
    )
)

(defiterationmethod (key-generic-aggregate :find-info) (i &key test)
    [:calliterationmacro (item index) :reduce (res nil)
	(let ((info (info-of item)))
	   (if test
	      (when (funcall test info i) (return index))
	      (when (flavor-equal info i) (return index))
	   )
	)
    ]
)
(defiterationmethod (key-generic-aggregate :count-info) (i &key test)
    [:calliterationmacro ((key info) index) :reduce
	(res 0)
	(if test
	   (if (funcall test info i) (1+ res) res)
	   (if (flavor-equal info i) (1+ res) res)
	)
    ]
)
(defiterationmethod (key-generic-aggregate :count-key) (k &key test)
    [:calliterationmacro ((key info) index) :reduce
	(res 0)
	(if test
	   (if (funcall test key k) (1+ res) res)
	   (if (flavor-equal key k) (1+ res) res)
	)
    ]
)

(defiterationmethod (key-generic-aggregate :is-info) (i) 
    (< 0 [:calliterationmethod :count-info i]))

(defiterationmethod (key-generic-aggregate :is-key) (k) 
    (< 0 [:calliterationmethod :count-key k]))

; -------------------- Iterators (page 80)

(defiterationmethod (key-generic-aggregate :iterate-key-method) (f)
    [:calliterationmacro (el in) :iterate
	(funcall f (key-of el) (info-of el) in)
    ]
)

; -------------------- Reduction (page 86)
(defiterationmethod (key-generic-aggregate :reduce-key-method) 
    (operation &key (unit-key nil unit-p) (unit-info nil unit-info-p))
    (if unit-p
       [:calliterationmethod :reduce-method
	   #'(lambda (res el in) 
	       (funcall operation (key-of res) (info-of res) 
				  (key-of el) (info-of el)   in))
           :unit (pair unit-key unit-inf)
       ]
       [:calliterationmethod :reduce-method
	   #'(lambda (res el in) 
	       (funcall operation (key-of res) (info-of res) 
				  (key-of el) (info-of el)   in))
       ]
    )
)

(defiterationmethod (key-generic-aggregate :reduce-key) 
    (operation &key (unit nil unit-p))

     [:calliterationmacro (el in) :iterate
	 (if unit-p
	    (setq unit (funcall operation unit (key-of el) in))
	    (progn (setq unit (key-of el)) (setq unit-p t))
	 )
     ]
)

(defiterationmethod (key-generic-aggregate :reduce-info) 
    (operation &key (unit nil unit-p))

     [:calliterationmacro (el in) :iterate
	 (if unit-p
	    (setq unit (funcall operation unit (info-of el) in))
	    (progn (setq unit (info-of el)) (setq unit-p t))
	 )
     ]
)


; -------------------- Find, Skip and Count (page 89)

(defmacro reduce-key-to-normal (key normal)
   `(defiterationmethod (key-generic-aggregate ,key) (property)
	[:calliterationmethod ,normal
	    #'(lambda (el in) 
		  (funcall property (key-of el) (info-of el) in))
	]
    )
)

(reduce-key-to-normal :find-property-key-method :find-property-method)
(reduce-key-to-normal :skip-property-key-method :skip-property-method)
(reduce-key-to-normal :count-property-key-method :count-property-method)

; -------------------- Existential and Universal Quantifiers

(reduce-key-to-normal :exists-key-method :exists-method)
(reduce-key-to-normal :forall-key-method :forall-method)

; -------------------- Order Dependent Operations

(defiterationmethod (key-generic-aggregate :max-info) ()
    (let (res-in res-key res-info)
        [:calliterationmacro ((key info) in) :iterate
	  (when (or (not res-in) 
		    (equal (flavor-compare info res-info) 'greater))
	     (progn (setq res-in in) 
		    (setq res-key key) 
		    (setq res-info info))
	  )
        ]
        (values res-info res-in)
    )
)

(defiterationmethod (key-generic-aggregate :min-info) ()
    (let (res-in res-key res-info)
        [:calliterationmacro ((key info) in) :iterate
	  (when (or (not res-in)
		    (equal (flavor-compare info res-info) 'less))
	     (progn (setq res-in in) 
		    (setq res-key key) 
		    (setq res-info info))
	  )
        ]
        (values res-info res-in)
    )
)

(compile-flavor-methods key-generic-aggregate)

