#|
   module      : lin-generic
   owner       : Franz Weber
   author      : Franz Weber
   explanation : generic functions of abstract linear aggregate like data types

      Linear  aggregates  have a fixed order of their component
      objects. Therfore it is possible to define

      1. a form of s-exp to represent them in lisp

      2. a position count (index of natural numbers) to identify
	 component objects. Position counts are positive natural
	 nunmbers starting with '1' !

      3. special traversing functions (first, next, previous, last)
         based on indices

      4. sorting of linear aggregates

      5. an iteration method (non virtually) for linear aggregates

      1. S-expressions
      ----------------

      S-expressions to denote linear aggregates are lisp lists. For
      instance:

      '(1 2 3) 

      denotes linear aggregate with '1' at position count 1,  '2' ...

      methods:

      (defmethod (generic-linear-aggregate :insert-s-exp) (s) ...)
      (defmethod (generic-linear-aggregate :move-in-s-exp) (s) ...)

      
      2. Position Counts
      ------------------

      Are positive natural numbers (1, 2 ...) to identify component
      objects in a linear aggregate.

      methods:

      (defmethod (generic-linear-aggregate :index-val) (p) ... )
	return index which belongs to the position count 'p'
	this method should be reimplemented by descendant classes
      (defmethod (generic-linear-aggregate :index-pos) (i) ...)
	return position count which belongs to the index 'i'
	this method should be reimplemented by descendant classes

      the next methods are the same as the corresponding methods based
      on indices:

      (defmethod (generic-linear-aggregate :get-by-pos) (p) ...)
      (defmethod (generic-linear-aggregate :set-by-pos) (p e) ... )
      (defmethod (generic-linear-aggregate :transfer-by-pos-in) (p e) ...)
      (defmethod (generic-linear-aggregate :transfer-by-pos-out) (p) ...)
      (defmethod (generic-linear-aggregate :modify-by-pos) (p modification) ...)


      3. Traversal
      ------------

      (defmethod (generic-linear-aggregate :first) () [:index-val 1])
	 return index of first component object
      (defmethod (generic-linear-aggregate :next) (i &key circular)
	 return index of component object which is the succesor of the
	 component object identified by indes 'i'. If circular is 't',
	 and i denotes the last object, return the index of the first
	 component object.
      (defmethod (generic-linear-aggregate :previous) (i &key circular)
	 reversse of next
      (defmethod (generic-linear-aggregate :last) () ...)
	 index of last component object

      
      4. Sorting
      ----------

      (defmethod (generic-linear-aggregate :sort)
	(&optional (list-implementation 'lis-list))

	sorts self, auxilliary list data structures are implemented
	by the type indicated by list-implementation, which must be
	a deescendant of class 'lis'.

	The implementation is a simple insertion sort.

	It returns self

      
      6. Iterators
      ------------

      (defiterationmethod (generic-linear-aggregate :iterate-normal-method)
	 (f) ...)

	 may be implemented on the basis of first, next, previous and
	 last.

      Iteration orders are 'from-first, 'from-last and 'dont-care.

     7. Mappings
     -----------

     (defiterationmacro (generic-linear-aggregate :map) (f) ...)

       return a new linear aggregate which is the application of f to each
       element of self.

       :map understands the parameters :result-type and :result-object
       map returns the object contained in result-object. The result
       object must be of a type, which is a sub type of
       linear-aggregate. If no result-object is given, a new object is
       created, which is of type result-type. The result-type must be
       a subtype of linear-aggregate. It is defaulted to the type of
       self.

       examples : (user input preceeded by '>')

      >  (setq self (make-instance 'lis-list :init-value '(2 3)))
      #<Instance lis-list BBFC4E>

      > [[:map (x) (* 2 x)] :dump]
      lis-list :
	4
	6
      #<Instance lis-list BC1DE6>

      >  [[:map :result-type 'lis-tab (x) (* 2 x)] :dump]
      lis-tab :
	4
	6
      #<Instance lis-tab BCA41E>

      > (setq w [:map :from [:find 3] :to [:find 3] (x) (* 3 x)])
      #<Instance lis-list BD311E>

      > [w :dump]
      lis-list :
	9
      #<Instance lis-list BD311E>

      > [:map :result-object w (x) (* 4 x)]
      #<Instance lis-list BD311E>

      > [w :dump]
      lis-list :
	8
	12
      #<Instance lis-list BD311E>



     (defiterationmacro (generic-linear-aggregate :filter) (f) ...)

     return list of elements, which return true on f, similar to map

     > (setq l (make-instance 'lis-list :init-value '(1 2 3 4 5)))
     #<lis-list 1 2 3 4 5>
     > [l :filter (x) (equal (mod x 2) 0)]
     #<lis-list 2 4>


   history     : 90.06.26
   errors      :
   bugs        : iterate-move-method seems not to be implemented
|#

(defflavor generic-linear-aggregate () (simple-aggregate))

; -------------------- constructors (page 57)

(defmethod (generic-linear-aggregate :clear) ()
    [:iterate :iteration-order 'from-last 'move (x i) [:remove i]]
)

(defmethod (generic-linear-aggregate :insert-s-exp) (s)
    (dolist (e s)
	[:insert e :to-pos `last]
    )
)

(defmethod (generic-linear-aggregate :move-in-s-exp) (s)
    (dolist (e s)
	[:move-in e :to-pos `last]
    )
)

(defmethod (generic-linear-aggregate :insert-aggregate) (s)
    [s :iterate (e)
	[:move-in e :to-pos `last]
    ]
)

(defmethod (generic-linear-aggregate :move-in-aggregate) (s)
    [s :iterate (i e)
	[:move  :to-pos `last :from-pos i :from s]
    ]
)

; -------------------- Access by Position Count (page 76)

; you are recommended to implement at least the next two methods

(defmethod (generic-linear-aggregate :index-val) (p) 
    (unless (eql p 0)
    (let ((c 0))
       [:iterate 
	   (el in)
	   (incf c)
	   (when (eql c p) (return in))
       ]
    )
)
)

(defmethod (generic-linear-aggregate :index-pos) (i) 
    (let ((c 0))
       [:iterate
	   (el in)
	   (incf c)
	   (when (flavor-equal i in) (return c))
       ]
       c
    )
)

(defmethod (generic-linear-aggregate :get-by-pos) (p) [:get-elem [:index-val p]])
(defmethod (generic-linear-aggregate :set-by-pos) (p e) [:set-elem [:index-val p] e])
(defmethod (generic-linear-aggregate :transfer-by-pos-in) (p e) 
    [:transfer-elem-in [:index-val p] e])
(defmethod (generic-linear-aggregate :transfer-by-pos-out) (p) 
    [:transfer-elem-out [:index-val p]])
(defmethod (generic-linear-aggregate :modify-by-pos) (p modification) 
    [:modify-elem [:index-val p] modification])

; -------------------- Selector Operations (page 77)
; all return an index
(defmethod (generic-linear-aggregate :first) () [:index-val 1])
(defmethod (generic-linear-aggregate :next) (i &key circular) 
    (let ((j [:index-pos i]))
       (incf j)
       (if (> j [:count])
	   (if circular [:first]  nil)
	   [:index-val j]
       )
    )
)
(defmethod (generic-linear-aggregate :previous) (i &key circular)
    (let ((j [:index-pos i]))
       (decf j)
       (if (< j 0)
	   (if circular [:last]  nil)
	   [:index-val j]
       )
    )
)
(defmethod (generic-linear-aggregate :last) () [:index-val [:count]])

; -------------------- Iterators (page 80)

(defiterationmethod (generic-linear-aggregate :iterate-normal-method) (f) 
; obey iteration-order part from to

    (if (eql iteration-order 'from-last)
       (progn
	  (if part
	      (progn
		 (setq to (cadr part))
		 (setq from (caddr part))
	      )
	      (unless from (setq from [:last]))
	  )
	  (when from
	     (let ((index from))
		 (loop
		     (funcall f [:get-elem index] index)
		     (when (eql index to) (return))
		     (setq index [:previos index])
		 )
	     )
	  )
       )
       (progn
	  (if part
	      (progn
		 (setq from (cadr part))
		 (setq to (caddr part))
	      )
	      (unless from (setq from [:first]))
	  )
	  (when from
	     (let ((index from))
		 (loop
		     (funcall f [:get-elem index] index)
		     (when (eql index to) (return))
		     (setq index [:next index])
		 )
	     )
	 )
      )
   )
)

; -------------------- Reduction (page 86)

; -------------------- Find, Skip and Count (page 89)

; -------------------- Existential and Universal Quantifiers

; -------------------- Order Dependent Operations

(defmethod (generic-linear-aggregate :sort) 
    (&optional (list-implementation 'lis-list))
    (let ((c (make-instance list-implementation)))
	[c :transfer-from self]
	(loop
	    (when [c :is-empty] (return))
	    [:insert [c :remove-min] :to-pos 'last]
	)
	[c :deallocate]
    )
    self
)
;(compile-flavor-methods generic-linear-aggregate)
