#|
   module      : simple-aggregate
   owner       : Franz Weber
   author      : Franz Weber
   explanation : general functions of abstract simple-aggregate like data types
                 See the book of Uhl/Endres
                 See the document 
		     'A Catalogue of Abstract Data Types in Flavors'
                 See the document 
		     'Aggregates'
  

        1. Iterations
        -------------
        Iterations are expressed by flavor macros. They all follow 
        a common scheme:

        '[' [<object>] :<iterator> <option>* <lambda> arbitrary ... ']'

        whereby a option may be a quoted symbol or a keyword value
        pair. A lambda must be a list, which may be arbitrary
        interpreted by the iteration macro. 
        Iterations may be exited by (return ...) .

        a) Standarditerations

        	For standard iterations the lamda may have one of the
        	following two forms:
        	(<elem> <index>) or ((<key> <info) <index)
        	Depending on the form of the lambda the 'method' or the
        	'key-method' is called. For Standarditeraions 'arbitrary'
        	means forms, which are passed as the body of the function
        	to apply on each element to the iteration methods.

        b) Reductions

        	Reductions have an additional part between lambda and
        	arbitrary:
        	... <lambda> <result-variable> <form>*
        	whereby <result-variable> may be
        	<symbol> or (<symbol> <init-expression>)
        	If an init-expression is given, this is taken as the unit
        	for the reduce method.

        c) Standard options are:
	
        move      -- alteration and deletion of the current element allowed
        recursive -- iterate recursive over sub aggregates 
                     (see pointer)
        reflexive -- iterate over self too (only when recursive is set)
	mark      -- same as recursive, but mark visited objects (in graphs)
	visited   -- set of marked objects, may be overwritten by a
	             set but is defaulted. It may be initialized by an 
		     non empty set. iteration is stopped at the
		     members of the set.
	result-type 	-- for special reductions like 'map'
			   type of the result, defaulted by type
			   of self.
	result-object	-- collection object for reduction, defaulted
			   by a newly created object of type result-type
		 

        iteration-order
            general : dont-care
            trees : pre-order, post-order, in-order
            lists : from-first from-last
        part
           general : 'whole-struct
           lists : `(range ,first-pos ,last-pos)
           trees : `(sub-tree ,pos)
        from
           lists, trees : pos
        to
           lists, trees : pos


	2. Iteration implementation support
        -----------------------------------

	Such iterations are defined by

        (defiterationmacro :<iterator> <object-name> <params> <forms>*)

        The options defined by the user are available by the symbol
        'keyword-options' in the forms of the macro. symbols are
        translated to :<symbol> t. As usual for flavor-macros, the
	object on which the iteration should work upon is available
	under  'object-name'. The params are an ordinary lambda list
	and serve to destructering bind the part '<lambda> arbitrary
	...' of a call to an iteration macro.

	A common case is, that the part 'arbitrary' are only forms of
	the body of the iteration. This is called a 'standard
	iteration' and is especially supported by the macro

	(defmacro defstandarditerationmacro (name &key method key-method) ...)

	For this kind of iterations only the name of the iteration and
	the names for the methods to call on this iteration must be
	supplied. 

	Methods for implementing such iterations have to distinguish
	between a lot of options. Therefore a special scheme to
	implement such iteration methods exists. It helps to keep up
	with the options for iterations. Methods for implementing
	itertions are defined by

	(defiterationmethod <method> <params> <forms> ...)

        The macro is a normal 'defmethod' but automatically defines
	the keyword parameters

	move recursive reflexive (iteration-order 'dont-care)
	part from to index-of-self

	In the body of such methods the macros 'calliterationmacro'
	and 'calliterationmethod' help to pass this bunch of options
	to other iterators and iterationmethods. An iterationmacro
	<im> is called  by

	[[<object>] :calliterationmacro [<options>] <lambda> <im> (forms)]

	Only those options must be mentioned, which shall be
	overwritten.  An iterationmethod <im> is called by

	[[<object>] :calliterationmethod <im> <args>]

        Only those options must be mentioned as arguments, which shall
	be overwritten.

	For the implementation of iteration methods it may be useful
	to further decompose this task, by implementing special
	options by special methods. This may easy the task to
	reimplement these methods in derived classes. Such a scheme is
	in general only necessary for the method 'iterate-method' of
	the standard iteration 'iterate', because most other methods
	are implemented relative to this method. The class
	'simple-aggregate' implements 'iterate-method' by the
	following forest:

	iterate-method
	    iterate-recursive-method
		-> iterate
		-> iterate-recursive-method
	    iterate-mark-method
		-> iterate
		-> iterate-mark-method
	    iterate-non-revursive-method
		iterate-move-method
		iterate-normal-method
		    -> iterate-move-method
	iterate-key-method
	    -> iterate
		
        The arrows indicat an implementation by a method mentioned
	elsewhere in the tree. The only method, which must be
	definitly implemented by a descendent class is the method
	'iterate-move-method'. However, out of efficency reasons
	descendant classes will also reimplement other parts of this
	tree.

	Note, that the function, which is passed to the
	iteration-methods have two arguments: the first parameter is
	the object, which is the object currently 'iterated upon', and
	the index of this object as the second argument. For 'key
	iterations' the function gets three arguments: the key, the
	info and the index of the object (in this order).


	3. Special Iterators
	--------------------

	iterate		simple standard iteration
	reduce		simple reduction (see  above)

	find-property   find first index for which the evaluation of
			the body returns true
        skip-property   same as find-property but, returning index
	                behind the first index matching the body.
        count-property  count the number of component objects matching
			the body
        exists		return boolean result, wether there exists a
			component object matching the body
        forall		return boolean result, wether all component
	                objects match the body


	'iteration methods'

	Some iteration methods don't need a body (because they define
	it themselves) but may use the iteration options:

	count				number of component objects
	is-empty			is there a component object
	find <elem> &key test           index of first equal object
	count-elem <elem> &key test     count objects equal to elem
	is-elem <elem> &key test        test wether there is an object
                                        equal to elem
	max                             return a list of the maximal 
					object and it's index
        min				same for minimal object


	4. Indices
	----------

	(virtualmethod (simple-aggregate :index-of) (pos))
		returns index  belonging to position pos


	i is an index, e a position :

	(virtualmethod (simple-aggregate :set-elem) (i e))
		set elem at index i to e
	(virtualmethod (simple-aggregate :get-elem) (i) )
		get component object at index i
	(virtualmethod (simple-aggregate :transfer-elem-out) (i))
		move out elem at index i leaving the structure
		of the aggregate as it is (The value at index i is
		undefined afterwards
	(virtualmethod (simple-aggregate :transfer-elem-in) (i e))
		move in elem e at index i (redefining the index)
	(defmethod (simple-aggregate :modify-elem) (i modification)
	    [:set-elem i (funcall modification [:get-elem i])]
	)


	5. Other Operations
	-------------------

	(defmethod (simple-aggregate :swap) (pos1 pos2)
		swaps objects on pos1 and pos2
		pos1 and pos2 are positions

	6. Miscellaneous
	----------------

	(defmethod (simple-aggregate :print-self) 
	    (stream prindepth &optional (indent 0)) ... )

	This method defines the print representations of simple
	aggregates. It is used by print and pprint. If you make own
	implementations, take this implementation as pattern. It obeys
	*print-length* *print-level* and *print-pretty* (as far as
	possible).

	:print-self is designed very modulary so that it is easily
	adapted to special needs. :print-self thinks of an object
	consisting of two parts: a lis of attributes and a simple
	aggregate of comonents. Everey object is printed in the
	following manner:

	"#<" (type-of self) 
	     <attribute-string> <attribute-lis> <component-lis> ">"

        :print-self-attribute-string prints the attribute string
        :print-self-attribute-lis prints the attribute lis
        :print-self-component-lis prints the component lis

	:attribute-string returns the attribute string of self
	    default ""
	:attribute-lis returns the attribute lis of self
	    default (make-instance 'lis-list)
	:component-lis returns the component lis of self
	    default self

        The function data-glo-print-lis may be used to print a lis of
	components according to the printing conventions.

	You may modify the behaviour by individually alter the methods
	mentioned above.

   history     : 90.06.26 Version 1
		 90.05.16 Franz Weber
			  print-self schema expanded by
				- component-lis
				- print-self-attribute-lis
				- print-self-component-lis
				- print-self-attribute-string
				- data-glo-print-lis
			  
   errors      :
   bugs        : parts are not considerd
                 'to-pos is not fully correct'
                 default implementations ofinsert-struct and move-struct are
                 not for all cases correct.
                 insert multiple is virtual
                 a lot of subclasses don't obey all
                 iteration variants
  todo	       : revisit passing of iteration parameters  
		 (by allow-other-keys ...)
		 move print-self in an extra flavor
|#

(mod-require "flavor/v4")
(mod-require "data/glo/v2/com")
(mod-require "data/glo/v2/cop")
(defflavor simple-aggregate () (comparison))
; -------------------- iteration macro
(defmacro defiterationmacro (name objectname params &rest forms)
    `(defflavormacro ,name ,objectname (&rest forms)
        (let (keyword-options options)
            (loop
                (when (and (listp (car forms)) 
                           (not (equal (car (car forms)) 'quote))) (return))
                (if (keywordp (car forms))
                   (progn
                      (setf keyword-options 
                          (append keyword-options  (list (car forms))))
                      (setf forms (cdr forms))
                      (setf keyword-options 
                          (append keyword-options  (list (car forms))))
                      (setf forms (cdr forms))
                    )
                    (progn
                      (setf options 
                          (append options  (list (cadr (car forms)))))
                      (setf forms (cdr forms))
                    )
                )
            )
            (setq keyword-options 
                (append keyword-options 
                    (mapcan
                        #'(lambda (x) 
                              (list
                                  (intern (string-upcase x) 'keyword)
                                  (list 'quote x)
                              )
                          )
                        options
                    )
                )
            )
            (destructuring-bind
                ,params
                forms
                `(block nil ,(progn ,@forms))
            )
        )
     )
)
(defmacro defstandarditerationmacro 
    (name &key method key-method )
   `(defiterationmacro ,name object (vars &rest forms)
        (if (listp (car vars)) ; key iteration form
           `[,object ,,key-method
                #'(lambda 
                      ,(if (cdr vars) 
                          (append (car vars) (cdr vars))
                          (append (car vars) (list (gensym)))) 
                      ,@forms
                  )
                ,@keyword-options
            ]
           `[,object ,,method
                #'(lambda 
                      ,(if (cdr vars) vars (append vars (list (gensym))))
                      ,@forms
                  )
                ,@keyword-options
            ]
        )
    )
)
(defstandarditerationmacro :iterate 
    :method :iterate-method 
    :key-method :iterate-key-method 
)
(defiterationmacro :reduce object (vars res &rest forms)
   (if (listp (car vars))
      (let ((item-name (gensym)))
         `[,object :reduce-method 
              #'(lambda 
                   ,(cons
                        (if (listp res) (car res) res)
                        (cons
                            item-name
                            (if (cdr vars) (cdr vars) (list (gensym))) 
                        )
                    )
                    (let ((,(caar vars) (key-of ,item-name))
                          (,(cadr vars) (info-of ,item-name)))
                    ,@forms)) 
              ,@(if (listp res)
                    (append
                        (list :unit (cadr res))
                        keyword-options
                    )
                    keyword-options
                )
          ]
       )
      `[,object :reduce-method 
           #'(lambda 
                ,(cons
                     (if (listp res) (car res) res)
                     (if (cdr vars) vars (append vars (list (gensym)))) 
                 )
                 ,@forms) 
           ,@(if (listp res)
                 (append
                     (list :unit (cadr res))
                     keyword-options
                 )
                 keyword-options
             )
       ]
   )
)
; ----- the following macros are to keep iteration method options
;       easy to maintain
(defmacro defiterationmethod (method params &rest forms)
   `(defmethod 
      ,method 
      ,(if (member '&key params)
           (append params
               '(move recursive reflexive (iteration-order 'dont-care)
                 part from to index-of-self mark visited
		 &allow-other-keys
                )
           )
           (append params
               '(&key move recursive reflexive (iteration-order 'dont-care)
                      part from to index-of-self mark visited
		      &allow-other-keys
                )
           )
       )
      ,@forms
    )
)
(defiterationmacro :calliterationmacro object-name (lambda name &rest forms)
   `(destructuring-bind
        (&key (move move)
              (recursive recursive)
              (reflexive reflexive)
              (index-of-self index-of-self)
              (iteration-order iteration-order)
              (part part)
              (from from)
              (to to)
	      &allow-other-keys
        )
        (quote ,keyword-options)
       [,object-name ,name
        :move move
        :recursive recursive
        :reflexive reflexive
        :index-of-self index-of-self
        :iteration-order iteration-order
        :part part
        :from from
        :to to
        ,lambda
        ,@forms
       ]
    )
)
(defflavormacro :calliterationmethod object-name (name &rest args)
    (let (iteration-args)
        (dolist (pair 
                    '((:move move)
                      (:recursive recursive)
                      (:reflexive reflexive)
                      (:index-of-self index-of-self)
                      (:iteration-order iteration-order)
                      (:part part)
                      (:from from)
                      (:to to)
                     )
                )
            (unless (member (car pair) args)
                (setq iteration-args (append pair iteration-args))
            )
        )
      `[,object-name ,name
        ,@args
        ,@iteration-args
       ]
    )
)
(defmacro if-struct (elem then &optional (else elem)) 
    `(if (typep ,elem 'simple-aggregate) ,then ,else)
)
; -------------------- Testing
(defflavormacro :print-self-macro object (stream prindepth &optional (indent 0))
   `(flavor-print-self ,object ,stream ,prindepth ,indent)
)

(defun flavor-print-self (object stream prindepth indent)
    (typecase object
	(simple-aggregate
	    [object :print-self stream prindepth indent]
	)
	(list
	   (flavor-print-self-list object stream prindepth indent)
	)
	(t
	   (format stream "~VT~A" indent
	      (with-output-to-string (s)
	      (write object :stream s :pretty nil
			     :level (when *print-level* 
					(- *print-level* (1+ prindepth)))
	      )
	      )
	   )
	)
    )
)

(defun flavor-print-self-list (object stream prindepth indent)
    (if (equal *print-level* prindepth)
       (format stream "#")
    (progn
    (format stream "~VT(" indent)
    (let ((need-blank nil) (count 0))
    (dolist (e object)
	(when (equal count *print-length*) 
	   (format stream " ...")
	   (return))
	(incf count)

	(if *print-pretty* 
	   (format stream "~%")
	   (when need-blank (format stream " "))
	)

        [e :print-self-macro 
	       stream 
	       (1+ prindepth)
	       (if *print-pretty* (+ indent 2) indent)]
    )
    )
    (format stream ")")
    )
    )
)

(defmethod (simple-aggregate :attribute-string) () nil)
(defmethod (simple-aggregate :attribute-lis) () (make-instance 'lis-list))
(defmethod (simple-aggregate :component-lis) () self)

; returns new count
(defun data-glo-print-lis (lis stream prindepth indent count)
    [lis :iterate (x)
	(when (equal count *print-length*) 
	    (format stream " ...")
	   (return))
	(incf count)
	(when *print-pretty* (format stream "~%"))
	[x :print-self-macro 
	   stream 
	   (1+ prindepth) 
	   (if *print-pretty* (+ indent 2) indent)]
    ]
    count
)

; returns new count
(defmethod (simple-aggregate :print-self-attribute-lis)
    (stream prindepth indent count)

    (data-glo-print-lis [:attribute-lis] stream prindepth indent count)
)

; returns new count
(defmethod (simple-aggregate :print-self-component-lis)
    (stream prindepth indent count)

    (data-glo-print-lis [:component-lis] stream prindepth indent count)
)

(defmethod (simple-aggregate :print-self-attribute-string) (stream)
    (let ((ss [:attribute-string]))
	(when ss (format stream " ~a" ss))
    )
)

(defmethod (simple-aggregate :print-self) 
	   (stream prindepth &optional (indent 0))

    (if (equal *print-level* prindepth)
       (format stream "#")
       (progn
	  (format stream "~VT#<~a" indent (type-of self))
	  [:print-self-attribute-string stream]
	  [:print-self-component-lis stream prindepth indent
	      [:print-self-attribute-lis stream prindepth indent 0]
	  ]
	  (format stream ">")
       )
    )
)

; -------------------- constructors (page 57)
(definitmethod (simple-aggregate) ())
(defmethod (simple-aggregate :swap) (pos1 pos2)
    (let* ((i [:index-of pos1]) ; see flavor position
           (j [:index-of pos2])
           (h [:transfer-elem-out i])
           (g [:transfer-elem-out j]))
        [:transfer-elem-in j h]
        [:transfer-elem-in i g]
    )
)
; -------------------- Operations based on indices (page 73)
(virtualmethod (simple-aggregate :index-of) (pos))
(virtualmethod (simple-aggregate :set-elem) (i e))
(virtualmethod (simple-aggregate :get-elem) (i) )
(virtualmethod (simple-aggregate :transfer-elem-out) (i)) 
(virtualmethod (simple-aggregate :transfer-elem-in) (i e)) 
(defmethod (simple-aggregate :modify-elem) (i modification)
    [:set-elem i (funcall modification [:get-elem i])]
)
; -------------------- Access by Position Count (page 76)
; -------------------- Selector Operations (page 77)
(defiterationmethod (simple-aggregate :count) ()
    [:calliterationmacro (x) :reduce (res 0)
        (1+ res) 
    ])
(defiterationmethod (simple-aggregate :is-empty) () 
    [:calliterationmacro (x) :reduce (res t) (return nil)])
(defiterationmethod (simple-aggregate :find) (e &key test)
     [:calliterationmacro (el in) :reduce (res nil)
         (when (if test (funcall test el e) (flavor-equal el e)) (return in))
     ]
)
(defiterationmethod (simple-aggregate :count-elem) (e &key test)
    [:calliterationmacro (el in) :reduce (res 0)
        (if (if test (funcall test el e) (flavor-equal el e)) 
            (1+ res)
            res
        )
    ]
)
(defiterationmethod (simple-aggregate :is-elem) (e &key test) 
    [:calliterationmethod :find e :test test])
; -------------------- Iterators (page 80)
(defiterationmethod (simple-aggregate :iterate-method) (f)
    (if recursive
        [:calliterationmethod :iterate-recursive-method f]
	(if mark
	   [:calliterationmethod :iterate-mark-method f]
	   [:calliterationmethod :iterate-non-recursive-method f]
	)
    )
)
(defiterationmethod (simple-aggregate :iterate-non-recursive-method) (f)
    (if move
        [:calliterationmethod :iterate-move-method f]
        [:calliterationmethod :iterate-normal-method f]
    )
)
(defiterationmethod (simple-aggregate :iterate-normal-method) (f)
    [:calliterationmethod :iterate-move-method f]
)
; must be implemented as iterationmethod
(virtualmethod (simple-aggregate :iterate-move-method) (f))

; hook for tree structures
; note, that not all aggregates understand the indices
(defiterationmethod (simple-aggregate :iterate-recursive-method) (f)
    (case iteration-order
        ((pre-order dont-care)
            (when reflexive (funcall f self index-of-self))
            [:iterate :move move (el in)
                [el :calliterationmethod
                    :iterate-recursive-method f 
                    :index-of-self  (cons (cons self in) index-of-self)
                    :reflexive t
                ]
            ]
        )
        (post-order
            [:iterate :move move (el in) 
                [el :calliterationmethod
                    :iterate-recursive-method f 
                    :index-of-self  (cons (cons self in) index-of-self)
                    :reflexive t
                ]
            ]
            (when reflexive (funcall f self index-of-self))
        )
        (in-order
            (let (b)
               [:iterate :move move (el in)
                   (if b
                       (when reflexive (funcall f self index-of-self))
                       (setq b t)
                   )
                   [el :calliterationmethod
                       :iterate-recursive-method f 
                       :index-of-self  (cons (cons self in) index-of-self)
                       :reflexive t
                   ]
               ]
            )
        )
    )
)

; hook for graph structures
; note, that not all aggregates understand the indices
(defiterationmethod (simple-aggregate :iterate-mark-method) 
    (f &key (visitied (make-instance 'set-hash)))
    (multiple-value-bind (index found) [visited :find-or-insert self]
    (unless found
    (case iteration-order
        ((pre-order dont-care)
            (when reflexive (funcall f self index-of-self))
            [:iterate :move move (el in)
                [el :calliterationmethod
                    :iterate-mark-method f 
                    :index-of-self  (cons (cons self in) index-of-self)
                    :reflexive t
                ]
            ]
        )
        (post-order
            [:iterate :move move (el in) 
                [el :calliterationmethod
                    :iterate-mark-method f 
                    :index-of-self  (cons (cons self in) index-of-self)
                    :reflexive t
                ]
            ]
            (when reflexive (funcall f self index-of-self))
        )
        (in-order
            (let (b)
               [:iterate :move move (el in)
                   (if b
                       (when reflexive (funcall f self index-of-self))
                       (setq b t)
                   )
                   [el :calliterationmethod
                       :iterate-mark-method f 
                       :index-of-self  (cons (cons self in) index-of-self)
                       :reflexive t
                   ]
               ]
            )
        )
    )
    )
    )
)

; -------------------- Reduction (page 86)
(defiterationmethod (simple-aggregate :reduce-method) 
    (operation &key (unit nil unit-p))
    (let ((result unit))
        [:calliterationmacro (el in) :iterate
              (if unit-p
                 (setq result (funcall operation result el in))
                 (progn (setq result el) (setq unit-p t))
              )
        ]
        result
    )
)
; -------------------- Find, Skip and Count (page 89)
(defstandarditerationmacro :find-property 
    :method :find-property-method 
    :key-method :find-property-key-method)
(defiterationmethod (simple-aggregate :find-property-method) (property)
    [:calliterationmacro (el in) :reduce  (res nil)
         (when (funcall property el in) (return in))
    ]
)
(defstandarditerationmacro :skip-property
    :method :skip-property-method
    :key-method :skip-property-key-method
)
(defstandarditerationmacro :count-property 
    :method :count-property-method
    :key-method :count-property-key-method)
(defiterationmethod (simple-aggregate :count-property-method) (property)
    [:calliterationmacro (el in) :reduce (res 0)
        (if (funcall property el in)  
            (1+ res)
            res
        ) 
    ]   
)
; -------------------- Existential and Universal Quantifiers
(defstandarditerationmacro :exists 
    :method :exists-method
    :key-method :exists-key-method)
(defiterationmethod (simple-aggregate :exists-method) (property) 
    [:calliterationmethod :find-property-method property])
(defstandarditerationmacro :forall 
    :method :forall-method
    :key-method :forall-method)
(defiterationmethod (simple-aggregate :forall-method) (property)
    [:calliterationmacro (el in) :reduce (res t)
        (unless (funcall property el in) (return nil))
        res
    ]
)
; -------------------- Order Dependent Operations
(defiterationmethod (simple-aggregate :max) ()
    (values-list
       [:calliterationmacro (el in) :reduce 
           (res (let ((any [:find-property (x) t]))
		    (list [:get-elem any] any)))
           (if (equal (flavor-compare el (car res)) 'greater)
               (list el in)
               res
           )
       ]
    )
)
#|
(defstandarditerationmacro :max-property
    :method :max-property-method
    :key-method :max-property-key-method
)
|#
(defiterationmethod (simple-aggregate :min) ()
    (values-list
       [:calliterationmacro (el in) :reduce 
           (res (let ((any [:find-property (x) t]))
		    (list [:get-elem any] any)))
           (if (equal (flavor-compare el (car res)) 'less)
               (list el in)
               res
           )
       ]
    )
)
#|
(defstandarditerationmacro :min-property
    :method :min-property-method
    :key-method :min-property-key-method
)
|#

; ----- mappings
(defstandarditerationmacro :map 
    :method :map-method 
    :key-method :map-key-method)
(defiterationmethod (simple-aggregate :map-method) 
    (f &key (result-type (type-of self)) 
	    (result-object (make-instance result-type) result-object-p))
    (when result-object-p [result-object :clear])
    (let ((insert-pos (if (eql iteration-order 'from-last) 'first 'last)))
       [:calliterationmacro (el in) :reduce (res result-object)
	   [res :move-in (funcall f el in) :to-pos insert-pos]
	   res
       ]
    )
)

(defmacro reduce-key-to-normal-simple-aggregate (key normal)
   `(defiterationmethod (simple-aggregate ,key) (property)
	[:calliterationmethod ,normal
	    #'(lambda (el in) 
		  (funcall property (key-of el) (info-of el) in))
	]
    )
)
; keep this consistent to key-gen

(reduce-key-to-normal-simple-aggregate :map-key-method :map-method)

(defstandarditerationmacro :filter
	:method :filter-method
        :key-method :filter-key-method)

(defiterationmethod (simple-aggregate :filter-method) 
    (f &key (result-type (type-of self)) 
	    (result-object (make-instance result-type) result-object-p))
    (when result-object-p [result-object :clear])
    (let ((insert-pos (if (eql iteration-order 'from-last) 'first 'last)))
       [:calliterationmacro (el in) :reduce (res result-object)
	   (when (funcall f el in) [res :move-in el :to-pos insert-pos])
	   res
       ]
    )
)
(reduce-key-to-normal-simple-aggregate :filter-key-method :filter-method)

(compile-flavor-methods simple-aggregate)
