#|
   module      : lis-list-readonly
   owner       : Franz Weber
   author      : Franz Weber/Olaf Mueller
   explanation : 

	concrete implementation of lis by lisp lists
	indices are the sublist beginning with the element pointed to

   history     : 92.02.18 Version 1
   errors      :
   bugs        : parts not implemented
		 normalize-remove/insert-pos
		 should be functions instead of methods !!
		 iterate method does not obey all variants
|#

(mod-require "data/lis/v2/lis")

(defflavor lis-list-readonly (content) (lis))

; --------------------

; kernel method
(definitmethod (lis-list-readonly) (&key init-value) (setf content nil)
    (setf content (append content init-value))
)

(defmethod (lis-list-readonly :cons) (h)
	(make-instance 'lis-list-readonly :init-value (cons h content)))

(defmethod (lis-list-readonly :copy-from) (source)
    (setq content
        [source :reduce :iteration-order 'from-last (x) (res nil)
            (cons x res)
        ]
    )
)

(defmethod (lis-list-readonly :copy) ()
    (make-instance 'lis-list-readonly :init-value content)
)

; returns the list without the last element; 
(defmethod (lis-list-readonly :butlast) () 
    (make-instance 'lis-list-readonly :init-value (butlast content))
)

(defmethod (lis-list-readonly :cdr) ()     
    (make-instance 'lis-list-readonly :init-value (cdr content))
)

(defmethod (lis-list-readonly :snoc) (h)   
    (make-instance 'lis-list-readonly :init-value (append content (list h)))
)


; -------------------- Operations based on indices (page 73)

(defmethod (lis-list-readonly :normalize-remove-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (first     nil)
	    (last      [:previous [:last]])
	)
	(case (car pos)
	    (at-index [:previous (cadr pos)])
	    (after    pos)
	    (first     nil)
	    (last      [:previous [:last]])
	)
    )
)


(defmethod (lis-list-readonly :index-of) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos (cdr pos) content))


(defmethod (lis-list-readonly :get-elem) (i) (car i))


(defiterationmethod (lis-list-readonly :iterate-non-recursive-method) 
    (f)
    ; obey iteration-order move part from to 
    ; move is transparently obeyed

     (if (eql iteration-order 'from-last)
	(progn
	   ; to should contain the whole list to iterate upon
	   ; from should contain the cdr of the first element
	   (if (consp part)
	       (progn
		   (setq from (cdadr part))
		   (setq to (car part))
	       )
	       (progn
	           (if to 
		       (setq to [:index-of to])
		       (setq to content))
		   (when from 
		       (setq from (cdr [:index-of from]))
		   )
	       )
	   )
	   (labels
	     ((rec (index)
		  (unless (eql index from) 
		      (rec (cdr index))
		      (funcall f (car index) index)
		  )
	      )
	     )
	     (when to (rec to))
	   )
	)
	(progn
	   ; from should contain whole list to iterate over
	   ; to should contain cdr of last element to iterate
	   (if (consp part)
	       (progn
		   (setq from (car part))
		   (setq to (cdadr part))
	       )
	       (progn
	           (if from 
		       (setq from [:index-of from])
		       (setq from content))
		   (when to 
		       (setq to (cdr [:index-of to]))
		   )
	       )
	   )
	    (when from
		(let ((n (cdr from)))
		    (loop
			(funcall f (car from) from)
			(when (equal n to) (return))
			(setq from n)
			(setq n (cdr from))
		    )
		)
	    )
	)
     )
)


; the following methods are only for implementation purposes,
; e.g. the mechanism filter. 
; they may not be used by normal users!!



; -------------------- Shift and Invert (page 132)

; this is an optimization
(defmethod (lis-list-readonly :invert) ()
    (setf content (list-nreverse content))
)

#| Methods for special checks

; -------------------- Parts (page 130)

(compile-flavor-methods lis-list)
; kernel method
(defmethod (lis-list :insert-s-exp) (e) 
    (setf content (append content (copy-list e))))

(defmethod (lis-list :clear) () (setf content nil) self)

; kernel method
(defmethod (lis-list :slice) (sub) 
    (setf content (car sub))
    (setf (cdadr sub) nil)
)



(defmethod (lis-list :move) 
    (from-pos &key (to-pos nil to-pos-p) (from nil from-p) &aux help)

    (unless to-pos-p (setf to-pos '(dont-care)))
    (unless from-p (setf from self))
    (setf help [from :move-out from-pos])
    [:move-in help :to-pos to-pos]
)
; kernel method
(defmethod (lis-list :move-out) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos
	 (prog1 (cadr pos) (setf (cdr pos) (cddr pos)))
	 (prog1 (car content) (setf content (cdr content)))
    )
)

; kernel method
(defmethod (lis-list :move-in) (e &key (to-pos nil to-pos-p))
    (unless to-pos-p (setf to-pos 'dont-care))
    (setq to-pos [:normalize-insert-pos to-pos])
    (if to-pos
	 (setf (cdr to-pos) (cons e (cdr to-pos)))
	 (setf content (cons e content))
    )
)

(defmethod (lis-list :normalize-insert-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
	(case (car pos)
	    (before [:previous (cadr pos)])
	    (after    pos)
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
    )
)

; kernel method
;(defmethod (lis-list :set-elem) (i e) (setf (car i) e))


(defmethod (lis-list-readonly :transfer-elem-out) (i) (car i))
; kernel method
;(defmethod (lis-list :transfer-elem-in) (i e) (setf (car i) e))
|#


; -------------------- Iterators (page 80)

(defiterationmethod (lis-list-readonly :map-method) 

    (f &key (result-type (type-of self)))

    (make-instance result-type :init-value
        (let ((insert-pos (if (eql iteration-order 'from-last) 'first 'last)))
           [:calliterationmacro (el in) :reduce (res nil)
	       (append  res (list (funcall f el in)))
           ]
        )   
    )
)
