#|
   module      : lis-sparse
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 

	implementations of sparse lists

	indices are the sublist beginning with the element pointed to
	plus an displacement

   history     : 90.06.26 Version 1
   errors      :
   bugs        : parts not implemented
		 normalize-remove/insert-pos
		 should be functions instead of methods !!
		 iterate method does not obey all variants
|#

(mod-require "data/lis/v2/lis")

(defflavor lis-sparse (content length) (lis) (:init-keywords :init-value :lis))

; --------------------

; kernel method
(definitmethod (lis-sparse) 
    (&key (lis (make-instance 'lis-list)) init-value) 
    [lis :clear]
    (setf content lis)
    [:move-in-s-exp init-value]
)

(defmethod (lis-sparse :count) () length)

; kernel method
(defmethod (lis-sparse :clear) () [content :clear] self)

; kernel method
(defmethod (lis-sparse :slice) (sub) 
    (let* ((fi (cadr sub))
	   (si (caddr sub))
	   (fii (car fi))
	   (fd  (cadr fi))
	   (sii (car si))
	   (sd  (cadr si))
	   (dis (- (+ fd (car [content :get-elem fii])) 1))
	  )
	(setf length (+ (car [content :get-elem sii]) sd))
	[content :slice `(part ,fii ,sii)]
	[content :iterate (x) (setf (car x) (- (car x) dis))]
    )
)

@@@@@

(defmethod (lis-sparse :move) 
    (from-pos &key (to-pos nil to-pos-p) (from nil from-p) &aux help)

    (unless to-pos-p (setf to-pos '(dont-care)))
    (unless from-p (setf from self))
    (setf help [from :move-out from-pos])
    [:move-in help :to-pos to-pos]
)

; kernel method
(defmethod (lis-sparse :move-out) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos
	 (prog1 (cadr pos) (setf (cdr pos) (cddr pos)))
	 (prog1 (car content) (setf content (cdr content)))
    )
)

; kernel method
(defmethod (lis-sparse :move-in) (e &key (to-pos nil to-pos-p))
    (unless to-pos-p (setf to-pos 'dont-care))
    (setq to-pos [:normalize-insert-pos to-pos])
    (if to-pos
	 (setf (cdr to-pos) (cons e (cdr to-pos)))
	 (setf content (cons e content))
    )
)

; -------------------- Operations based on indices (page 73)

(defmethod (lis-sparse :normalize-remove-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (first     nil)
	    (last      [:previous [:last]])
	)
	(case (car pos)
	    (at-index [:previous (cadr pos)])
	    (after    pos)
	    (first     nil)
	    (last      [:previous [:last]])
	)
    )
)

(defmethod (lis-sparse :normalize-insert-pos) (pos)
    ;faulty
    (if (symbolp pos)
	(case pos
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
	(case (car pos)
	    (before [:previous (cadr pos)])
	    (after    pos)
	    (dont-care nil)
	    (first     nil)
	    (last      [:last])
	)
    )
)

(defmethod (lis-sparse :index-of) (pos) 
    (setq pos [:normalize-remove-pos pos])
    (if pos (cdr pos) content))

; kernel method
(defmethod (lis-sparse :set-elem) (i e) (setf (car i) e))
(defmethod (lis-sparse :get-elem) (i) (car i))
(defmethod (lis-sparse :transfer-elem-out) (i) (car i))
; kernel method
(defmethod (lis-sparse :transfer-elem-in) (i e) (setf (car i) e))

; -------------------- Iterators (page 80)

(defiterationmethod (lis-sparse :iterate-non-recursive-method) 
    (f)
    ; obey iteration-order move part from to 
    ; move is transparently obeyed

     (if (eql iteration-order 'from-last)
	(progn
	   ; to should contain the whole list to iterate upon
	   ; from should contain the cdr of the first element
	   (if (consp part)
	       (progn
		   (setq from (cdadr part))
		   (setq to (car part))
	       )
	       (progn
	           (unless to (setq to content))
		   (when from (setq from (cdr from)))
	       )
	   )
	   (labels
	     ((rec (index)
		  (unless (eql index from) 
		      (rec (cdr index))
		      (funcall f (car index) index)
		  )
	      )
	     )
	     (when to (rec to))
	   )
	)
	(progn
	   ; from should contain whole list to iterate over
	   ; to should contain cdr of last element to iterate
	   (if (consp part)
	       (progn
		   (setq from (car part))
		   (setq to (cdadr part))
	       )
	       (progn
	           (unless from (setq from content))
		   (when to (setq to (cdr to)))
	       )
	   )
	    (when from
		(let ((n (cdr from)))
		    (loop
			(funcall f (car from) from)
			(when (equal n to) (return))
			(setq from n)
			(setq n (cdr from))
		    )
		)
	    )
	)
     )
)

; -------------------- Parts (page 130)

(compile-flavor-methods lis-sparse)
