#|
   module      : lis
   owner       : Franz Weber
   author      : Franz Weber
   explanation : abstract data type lis

	A list (lis) is a linear aggregate with the usual meanings of
	lists.

	1. Multiple Iterations
	----------------------

	For lists it is possible to iterate simulatneously over
	multiple lists. The iteration stops as soon as the end of the
	shortest of the lists is reached.

	Syntax:

	(multiple-iterate ( (<elem>  [ <index> ] <list>) * ) <form>* )

	2. Generic Implementations of Virtual Aggregate Methods
	-------------------------------------------------------

	(defmethod (aggregate :skip) (e) ... )
	(defmethod (lis :skip-property) (property) ... )
	(defmethod (lis :equal) (right) ... )
	(defmethod (lis :compare) (right) ... )
	    lexicographic comparison
	

	3. List Operations
	------------------

	(defmethod (lis :shift) (&key (by 1))
	    shifts circulary elements 'by' positions to the right.
	    Negative numbers are allowed (left shift).
	(defmethod (lis :invert) ()
	    invert order of list
	    return self

	runs:  (not implemented yet, but interface defined)
	A run is a sublist of equal elements.

	(defmethod (lis :run) (e min-length)
	    return index of first run  with elements equal to e
	    with minimal length 'min-length'
	(defmethod (lis :longest-run) (e)
	    return index of longest run of elements equal to e
	(defmethod (lis :max-run)
	   retrun index of longest run of any element.

	(defmethod (lis :car) () [:get-elem [:first]])
	   returns first component object of list
        (defmethod (lis :mapcar) (from modification)
	  make self to a list, which consists of the elements of from
	  (iin the same order), but each element is modified by the
	  function 'modification'

	  obsolete by :map in lin-gen @@@@@@@@@@@@@@@@@:wq


	4. Positions and Parts
	----------------------

	There are the following insert positions:

	    'first
	    'last
	    `(after ,<index>)
	    `(before ,<index>)

	There are the following remove positions:

	    'first
	    'last
	    `(at-index ,<index>)
	    `(after ,<index>)

        There are the following parts:

	   'whole-struct
	   `(,<first-index> ,<last-index>)

	It is strongly recommended to implement the methods

	(defmethod (lis :normalize-insert-pos) (p))
	(defmethod (lis :normalize-remove-pos) (p))

	for each concrete implementation of lis.


   history     : 90.06.26 Version 1
   errors      :

	shift is wronlgy implemented

   bugs        : there should be an implementation, which
		 may be initialized by an arbitrary sequence
		  
                 there should be component implementations for
		 all list variants

		 multiple iterations should be lifted to linear
		 aggregates.

		 mapcar  should be a kind of iterator


|#

(mod-require "data/glo/v2/lin")

(defflavor lis () (linear-aggregate)) 

; -------------------- Double Iterations (page 132)
(defmacro multiple-iterate
    (header &rest forms)

    (let ((vars (mapcar 
		    #'(lambda (x)
			  (if (cddr x)
			      x
			      (list (car x) (gensym) (car (last x)))
			  )
		      ) 
		    header)))
       `(progn
	    ; generate simultaneous initialization
	    (let ,(mapcar #'(lambda (v) (list (cadr v) `[,(caddr v) :first]))
		       vars
		  )
		 
	       (loop
		   ; end test
		   (unless 
		       ,(cons 
			    'and 
			    (mapcar
				#'(lambda (x)
				      (cadr x)
				  )
				vars
			    )
			)
		       (return))
		   
		   ; execute body
		   (let ,(mapcar 
			      #'(lambda (v) 
				    (list 
					(car v) 
					`[,(caddr v) :get-elem ,(cadr v)]))
			      vars 
			  )   
		       ,@forms
		   )
		   ; switch to next position
		   (setf ,@(mapcan
			      #'(lambda (v) 
				    (list 
					(cadr v) 
					`[,(caddr v) :next ,(cadr v)]))
			      vars 
			  )   
			 

		   )
	       )
	    )
	)
     )
)

; -------------------- constructors (page 57)

(defmethod (lis :copy-from) (from)

     [:clear]
     [from :iterate
	 (el)
	 [:move-in el :to-pos 'last ]
     ]
     self
)

; -------------------- Selector Operations (page 77)

(defmethod (lis :skip) (e)
    [:find-porperty (x) (not (flavor-equal x e))]
)
    ; returns an index

(defmethod (lis :skip-property) (property)
    [:reduce (el in) (res nil)
          (unless (funcall property el in) (return in))
    ]
)

(defmethod (lis :equal) (right)

    (and
	(equal [:count] [right :count])
	(let ((b t))
	   (multiple-iterate
	       ((r  right) (s self))
	       (unless (flavor-equal r s) (progn (setf b nil) (return)))
	   )
	   b
	)
    )
)

; -------------------- Order Dependent Operations

(defmethod (lis :compare) (right)
     (let ((c 'equal))
	 (multiple-iterate
	     ((s self) (r right))
	     (case (flavor-compare s r)
		 ('less (setf c 'less) (return))
		 ('greater (setf c 'greater) (return))
	     )
	 )
	 (if (equal c 'equal)
	     (flavor-compare [:count] [right :count])
	     c
	 )
     )
)

; -------------------- Shift and Invert (page 132)

(defmethod (lis :shift) (&key (by 1))

    (let ((c [:count]))
	(loop
	    (when (<= (- c by) 0) (return))
	    [:move [:first] [:last]]
	    (decf c)
	)
    )
)

(defmethod (lis :invert) ()
    (let
	((f [:first])
	 (l [:last])
	 f1 l1)

	 (loop
	     (when (equal f l) (return))
	     [:swap `(at-index, f) `(at-index ,l)]
	     (setf f1 [:next f])
	     (setf l1 [:previous l])
	     (when (equal f l1) (return))
	     (setf f f1)
	     (setf l l1)
	 )
    )
    self
)

; -------------------- Runs (page 136)

(defmethod (lis :run) (e min-length)
    (may be generic implemented)
)
(defmethod (lis :longest-run) (e)
    (may be generic implemented)
)
(defmethod (lis :max-run)
    (may be generic implemented)
)

; -------------------- Positions (page 129)

; 'first 'last index = '(after ,index) '(before ,index)
; implementors : be carful in case of the empty lis
;;(defmethod (lis :normalize-insert-pos) (p))

; -------------------- Parts (page 130)

; 'whole-struct
; `(,first ,last)

; ------------------- Lisp Operators (new invention

(defmethod (lis :car) () [:get-elem [:first]])

(defmethod (lis :mapcar) (from modification)

    [from :reduce 
	#'(lambda (x y) 
	      [x :insert (funcall modification y) :to-pos [x :insert-last]]
	      x
	  ) 
	#'(lambda (x) t)
	#'(lambda (x) nil)
	:unit [:clear]
    ]
)
(compile-flavor-methods lis)
