#|
   module      : simple-fix-tree
   owner       : Franz Weber
   author      : Franz Weber
   explanation : 
   
	This ADT implements trees. Trees are seen as a hierarchy of
	simple aggregates. This hierarchy is the content of the tree.
	the nodes are the individual simple aggregates. Although part
	of the tree, the nodes are objects in their own right. They
	may be manipulated individually. You may use pointers to build
	the leaves of the tree.

	This ADT differs from the trees described in the book of 
	Uhl/Schmidt. This data type doesn't allow unvalid sons, which
	makes it more comfortable for formula or syntax trees.

	(defflavor simple-fix-tree (content) (simple-aggregate copies)
	   (:settable-instance-variables content)
	       (:gettable-instance-variables content)
	       (:init-keywords :content)
	)

	1. Positions  & Parts
	---------------------

	There are the following (remove) positions:

	'root
	`(at-index ,<index>)
	'dont-care
	`(son-of ,<node-index> ,<node-pos>)
	    which means the son of node with node-index at
	    position node-pos within that node

        Implementation hint:

	    The generic index implementation is
	    'root or
	    (list of (fathers node-index))

	Parts :

	'whole-struct
	`(sub-tree ,<sub-tree>)
	   where <sub-tree> is index of a node


	2. Generic Implementations
	--------------------------

	(defmethod (simple-fix-tree :copy-from) (from) ... )
	(defmethod (simple-fix-tree :deallocate) () ... )

	(defmethod (simple-fix-tree :index-of) (pos) ... )
	(defmethod (simple-fix-tree :set-elem) (i e) ... )
	(defmethod (simple-fix-tree :get-elem) (i) ... )
	(defmethod (simple-fix-tree :transfer-elem-in) (i e) ... )
	(defmethod (simple-fix-tree :transfer-elem-out) (i) ... )
	(defiterationmethod (simple-fix-tree :iterate-method) ... )

	(defmethod (simple-fix-tree :slice) (part)

	3. Specific Tree Operations
	---------------------------

        (defmethod (simple-fix-tree :root) () 'root)
		return index of root node
	(defmethod (simple-fix-tree :nr-of-sons) (i) [:get-elem i] :count])
		return number of sons of node with index i
	(defmethod (simple-fix-tree :son) (i nr)
		return index of son number 'nr' of node with index i
	(defmethod (simple-fix-tree :father) (i) (cdr i))
		return index of father of node with index i
	(defmethod (simple-fix-tree :get-father) (i)
		return the father node
	(defmethod (simple-fix-tree :number-of-son) (father son)
		return the sons number in respect to all sons of
		father (each given as indices).
	(defmethod (simple-fix-tree :find-father) (son)
		return values
			[:father son]
			[:nr-of-son father son]
	(defmethod (simple-fix-tree :compare) (right)
		compare in prefix order
		may be implemented erroneously


	not implemented yet: 

	(defmethod (simple-fix-tree :rotate) (pos))


   history     : 90.08.14 Version 1
   errors      :
   bugs        : slice doesn't deallocate nodes thrown away
|#

(mod-require "data/glo/v2/simple-aggregate")
(mod-require "data/glo/v2/cop")

(defflavor simple-fix-tree (content) (simple-aggregate copies)
    (:settable-instance-variables content)
    (:gettable-instance-variables content)
    (:init-keywords :content)
)

; ----- copies
(defmethod (simple-fix-tree :copy-from) (from)
    (setq content [[from :content] :copy])
    self
)

(defmethod (simple-fix-tree :deallocate) ()
     [content :deallocate]
     (setq content nil)
)

; ----- simple aggregate

#|
Positions are :
'root
`(at-index ,index)
`dont-care
`(son-of ,node-index ,nod-pos)

a generic index is
'root
(list of (fathers node-index))

|#

(defmethod (simple-fix-tree :index-of) (pos) 
    (cond
	((member pos '(root dont-care)) 
	    'root
	)
	((eql (car pos) 'at-index)
	    (cadr pos)
	)
	((eql (car pos) 'son-of) 
	    (cons
		(cons
		    [:get-elem (cadr pos)]
		    (caddr pos)
		)
		(cadr pos)
	    )
	)
	(t (error "data/tre/v2/simple-tre: unkown index ~a ~%" pos))
    )
)

(defmethod (simple-fix-tree :set-elem) (i e)
    (if (eql i 'root)
	(progn
	   [content :deallocate]
	   (setq content e)
	)
	[(caar i) :set-elem (cdar i) e]
    )
)

(defmethod (simple-fix-tree :get-elem) (i)
    (if (eql i 'root)
	content
	[(caar i) :get-elem (cdar i)]
    )
)

(defmethod (simple-fix-tree :transfer-elem-in) (i e)
    (if (eql i 'root)
	(setq content e)
	[(caar i) :transfer-elem-in (cdar i) e]
    )
)

(defmethod (simple-fix-tree :transfer-elem-out) (i)
    (if (eql i 'root)
	(prog1 content (setq content nil))
	[(caar i) :transfer-elem-out (cdar i)]
    )
)

(defiterationmethod (simple-fix-tree :iterate-method) 
    (f)
   [content :calliterationmethod :iterate-method 
       f :recursive t :reflexive t :index-of-self 'root]
)

; -------------------- constructors

; parts are 'whole-struct and `(sub-tree ,sub-tree)
(defmethod (simple-fix-tree :slice) (part) 
    (unless (eql part 'whole-struct)
	(setq content (cadr part))
    )
)

; -------------------- specific tree operators (page 164)

(defmethod (simple-fix-tree :root) () 'root)

(defmethod (simple-fix-tree :nr-of-sons) (i) [:get-elem i] :count])

(defmethod (simple-fix-tree :son) (i nr)
    (let ((elem [:get-elem i]))
	[:index-of `(son-of ,i [[:get-elem i] :index-val nr])]
    )
)

(defmethod (simple-fix-tree :father) (i) (cdr i))

(defmethod (simple-fix-tree :get-father) (i)
    (caar i) ; = [:get-elem [:father i]]
)

(defmethod (simple-fix-tree :number-of-son) (father son)
    [[:get-elem father] :index-pos [[:get-elem father] :find son]]
)

(defmethod (simple-fix-tree :find-father) (son)
    (let ((father [:father son]))
	(values father [:nr-of-son father son])
    )
)

; rotations not implemented yet

(defmethod (simple-fix-tree :rotate) (pos) (error "not implemented"))

(defmethod (simple-fix-tree :compare) (right)
    (if (typep right 'simple-fix-tree)
        [content :compare [right :content]]
	(flavor-compare (type-of self) (type-of right))
    )
)

(defmethod (simple-fix-tree :equal) (right)
    (or (eql self right)
	(and (typep right 'simple-fix-tree)
             [:content |:equal [right :content]]
	)
    )
)

; ----- test

(defmethod (simple-fix-tree :component-lis) () (make-instance 'lis-list))
(defmethod (simple-fix-tree :attribute-lis) ()
    (make-instance 'lis-list
	:init-value (list content)
    )
)
