 #|
   module      : flavor
   owner       : Franz Weber
   author      : Franz Weber
   explanation : flavor conventions

	1. short terms for send

	   a) [object :message params ...]

	      means

	      (send object :message params ...)

           b) [:message params ...]

	      means

	      [self :message params ...]
	   
	   c) [object :message params ... | :message1 params ... | ...  ]

	      means


	      [[[ object :message parmas ...] :message1 params ... ] ... ]
	
	2. special method declarations

	   a) (virtualmethod <method description> <params>)

	      declares a method which has to be implemented by
	      descendant classes. This enables 'abstract classes'.

	   b) (hidemethod <method description>)

	      hides a method for a class (to forbid certain inherited
	      messages

           c) (definitmethod (class [<method combinator]) <params> <forms>)
	      
	      allows the definition of a method, which is called
	      at the creation of an object. Note, that only keyword
	      parameters are allowed, which where declared as
	      'init-keywords' in the flavor definition.

	3. flavor macros

	   (defflavormacro :name objectname params forms)

	   declares a macro, which may be called as

	   [object :name ...] or
	   [:name           ]

	   whereby the lexical variable objectname is bound to the object
	   implied by the semantic of []

	4. flavor speedup

	   The function flavor-unsecure disables the 'cleanup'
	   during make-instance. This results in a speed up of
	   factor 6 to 10 for make-instance. However no new defflavors
	   or defmethods are handled correctly. You have to use
	   compile-flavor-methods for each flavor before you
	   use flavor-unsecure. 'flavor-secure' undoes the effects of
	   'flavor-unsecure'.

        5. Mixins
	   One may 'lift' methods of components of a flavor to the
	   instance itself by 

	   (flavor-mixin <flavor> <component> <method-name> *)

   history     :
  
	90.06.12
	90.07.12 short form for self introduced
		 [:f ...] means [self :f ... ]
	90.08.15 speed up routines
  
   bugs        :
	
	short terms don't work in debug mode

	dispatch of [ and interpretation of arguments is not
	separated. This causes the interpretation of arguments
	allways at read time which may be erroneous.

   errors      :
|# 

(export 
    '(definitmethod 
      virtualmethod 
      hidemethod 
      *flavor-macros* 
      defflavormacro
      undefflavormacro
      flavor-send
     )
)

; -------------------- special method declarations
(defmacro definitmethod (message lambda-list &rest forms)
    `(defmethod 
        ,(append message (list :init))
	(args)
	(destructuring-bind
	    ,lambda-list
	    args
	    ,@forms
	)
    )
)

(defmacro virtualmethod (&rest dummy) nil)

(defmacro hidemethod  (message)
    `(defmethod ,(list (first message) :before (car (last message))) (&rest args)
        (error "method may not be implemented by descendant class")
    )
)

(defparameter *flavor-macros* nil)

; -------------------- short terms for send

(defmacro flavor-send-1 (&rest l)

    (setq l (if (keywordp (car l)) (cons 'self l) l))
    (if (member (cadr l) *flavor-macros*)
	(cons (cadr l) (cons (car l) (cddr l)))
	(cons 'send l)
    )
)

(defmacro flavor-send (&rest l)
    (let ((res nil))
	(dolist (e l)
	    (cond 
		((equal e #\|)
		    (setq res (list (cons 'flavor-send-1 res)))
		)
		(t (setq res (append res (list e))))
	    )
	)
	(cons 'flavor-send-1 res)
    )
)

(set-macro-character
    #\[
    #'(lambda (s c &aux l m)
	  (setq l nil)
	  (loop
	      (setq m (read s nil (values) t))
	      (when (eql m #\]) (return))
	      (setf l (append l (list m)))
	  )
	  (cons 'flavor-send l)
      )
      nil
)

(set-macro-character
    #\|
    #'(lambda (s c &aux l)
	  #\|
      )
      nil
)

(set-macro-character
    #\]
    #'(lambda (s c &aux l)
	  #\]
      )
      nil
)

; -------------------- flavor macros

(defmacro defflavormacro (name objectname params &rest forms)


   `(progn
       (setq *flavor-macros* (cons ,name *flavor-macros*))
       (defmacro ,name ,(cons objectname params) ,@forms)
    )
)

(defmacro undefflavormacro (name) 
    (setq *flavor-macros* (delete name *flavor-macros*))
    nil
)

; -------------------- fast flavors

(defparameter  *flavors-cleanup-flavor-function* 
    (symbol-function 'flavors::cleanup-flavor))

(defvar *flavors-cleanup-flavor-void* #'(lambda (&rest args) ))

(defun flavor-secure ()
    (setf (symbol-function 'flavors::cleanup-flavor)
	*flavors-cleanup-flavor-function*
    )
)

(defun flavor-unsecure ()
    (setf (symbol-function 'flavors::cleanup-flavor)
	*flavors-cleanup-flavor-void*
    )
)

; -------------------- Mixins

(defmacro flavor-mixin (flavor component &rest methods)
   `(progn
	,@(mapcar
	      #'(lambda (x)
		    `(defmethod (,flavor ,x) (&rest args)
			(apply 'send ,component ,x args)
		     )
		)
	      methods
	  )
    )
)
