;;;;-----------------------------------------------------------;;;;   
;;;;                 package LAZY-SENDING                      ;;;;
;;;;-----------------------------------------------------------;;;;
(in-package 'LAZY-SENDING)
 (import 'flavor-internals::send)
(import 'user::kwote)
(export '(reset-lazy-send eval-lazy-send lazy-send))
 
;;; Permits a lazy evaluation of SEND messages (flavor-package)
;; USAGE: if a flavor-method sends a msg to another instance
;;        but is not interested in the result it could send this 
;;        message lazy by LAZY-SEND. This is especially useful if the
;;	  other instance sends again so that the call-stack grows. 
;;	  It is possible to give lazy-send a list of receivers of the 
;;        message and a test. The message will be send to receivers
;;	  in turn until the result of send satisfies test.
;;        The lazy-send mechanism must be initialized with RESET-LAZY-SEND,
;;	  then making an initial (LAZY-SEND init-msg) and calling
;;        (EVAL-LAZY-SEND) which evaluates the lazy-msgs
;;	  last-in-first-send until there are no more.
	                              
(defvar *lazy-send-stack* nil)

(defun reset-lazy-send ()
    (setq *lazy-send-stack* nil) )
    
(defun eval-lazy-send ()
   "evaluates the lazy-send-stack until empty"
   (do ()((null *lazy-send-stack*) 'eval-lazy-send)
         (eval (pop *lazy-send-stack*))))
         
(defmacro lazy-send (to (method &rest params) 
                       &key ((:until tst) '(quote and)) )
"pushes a lazy-send-msg on the stack
 TO : the receiver or the list of receivers. TO is evaluated
 (METHOD PARAMS) : the method to be applied with its parameters
                   METHOD is not evaluated, PARAMS are evaluated
 TST : relevant only if to is a list. The message is send to the
       members of TO until TST is true for the result of the message.
       TST is not evaluated and should be a one-argument-function    "
                             
  ;;; pushes the list (.._eval means that this form is evaluated in the
  ;;;                    calling environment before pushing )
  ;;; (send-all 'to_eval 'tst 'method 'p1_eval ... 'pn_eval) 
  ;;; i.e. all arguments of a pushed send(-all) are evaluated before the push
 
   `(push (list* (quote send-all) (kwote ,to) (quote ,tst)
                 (quote (quote ,method))
                 (mapcar #'kwote (list ,@params)))
          *lazy-send-stack*) )

(defun send-all (tolist tst method &rest params)
  (let ((success nil))
    (do ((still-to (if (listp tolist) tolist (list tolist))
                   (cdr still-to)))
        ((or (null still-to) success)) 
        (setq success
              (eval (list tst (list* 'flavor-internals::send
                                     (kwote (car still-to))
                                     method (mapcar #'kwote params)))))))) 

(in-package 'user)
(import '(lazy-sending::reset-lazy-send 
          lazy-sending::eval-lazy-send 
          lazy-sending::lazy-send))
 
;;;;-----------------------------------------------------------;;;;   
;;;;                END package LAZY-SEND                      ;;;;
;;;;-----------------------------------------------------------;;;;      


