;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;	   M O D U L E   L I S P - E X T E N S I O N		;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;           EXTENSIONS to LISP of GENERAL use 		;;;;
;;;;             (chapters according to Steele) 		;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;           AUTHOR: Paul Weber, FZI, Karlsruhe 		;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(provide 'lisp-extension)

(in-package 'LISP-EXTENSION)
(use-package 'windows)  ;; only used for format-table

;; global variables
(export 
  '(*bin-extension*
      ; file-type of compiled lisp (may be sbin, lbin, fasl
    *at-console*
      ; is t if user is sitting at console of the lisp-host
      ; this var has to be checked before generating windows
  ))
	                
;; functions	                
(export
  '(boolean				; Ch.  4: Type specifier
    nlambda				; Ch.  5: Program Structure
    true false neq                      ; Ch.  6: Predicates
    kwote case-eval setv                ; Ch.  7: Control Structure
    string-symbol format-symbol         ; Ch. 10: Symbols
    flatten permutate length< lenght>   ; Ch. 15: Lists
    pushl
    cat					; Ch. 18: Strings
    format-table stop-until-enter	; Ch. 22: Input/output
    save-setq save-defun save-defstruct 
    fil-add-suffix fil-default-suffix   ; Ch. 23: File system interface
    fil-replace-suffix fil-is-newer 
    path-conc full-pathname 
    msecs			        ; Ch. 25: Miscellaneous features
  ))


;;;;-----------------------------------------------------------;;;;   
;;;;			0. VARIABLES                           ;;;;
;;;;-----------------------------------------------------------;;;;   

(defvar *bin-extension* ; defines the file-type of compiled-files
    (case (read-from-string (machine-type))
                      (Sun-4         "sbin")
                      ((SUN-2 Sun-3) "lbin")
                      (t             "fasl"))
   ) 
                      
(defvar *at-console* ; is t if lisp was started from console else nil
  ; for starting the lips-windows this var should be t
  ;;; code MUST still be written !!!
   ;(y-or-n-p "Are you sitting at the console of the Lisp-Host")
   ; (shell "if (`tty` == /dev/console) then
   ;        echo t else
   ;        echo nil
   ;        endif")
  t)  

;;;;-----------------------------------------------------------;;;;   
;;;;		     4. TYPE SPECIFIER                         ;;;;
;;;;-----------------------------------------------------------;;;;   

(deftype boolean () `symbol)

;;;;-----------------------------------------------------------;;;;   
;;;;		    5. PROGRAM STRUCTURE                       ;;;;
;;;;-----------------------------------------------------------;;;;   

(defmacro nlambda (f &rest args)
 "prevents evaluation of arguments of f"
    `(,f ,@(mapcar `(lambda (x) `(quote ,x)) args))
)

;;;;-----------------------------------------------------------;;;;   
;;;;			6.  PREDICATES                         ;;;;
;;;;-----------------------------------------------------------;;;;   

(defmacro true  (&rest x)
"returns always t, arbitrary many arguments" 
  (declare (ignore x))
  t)

(defmacro false (&rest x)
"returns always nil, arbitrary many arguments"
  (declare (ignore x))
  nil)
  
(defmacro neq (x y)
"equivalent to (not (eq x y))"
  `(not (eq ,x ,y))
 )

;;;;-----------------------------------------------------------;;;;   
;;;;                7. CONTROL STRUCTURE                       ;;;;
;;;;-----------------------------------------------------------;;;;

(defmacro kwote (form)
 "returns form quoted, like Franzlisp-Kwote"
  `(list 'quote ,form))
      ; as function body is: (list 'quote form)

(defmacro case-eval (keyform &rest forms)
 "like case, but evaluates all the keys in its keylist"
   (let ((new-var (gensym)))
      `(let ((,new-var ,keyform))
	 (cond
	    ,@(mapcar
	       #'(lambda (e)
		  (cons
		     (cond
			((or (eql 'otherwise (car e))
			     (eql 't (car e)))
			   `t
			)
			((symbolp (car e)) 
			   `(eql ,new-var ,(car e)) )
			((cdar e)
			   `(member ,new-var (list ,@(car e))))
			(t `(eql ,new-var ,(caar e)))
		     )
		     (cdr e)
	       )  )
	       forms
   )  )  )  )
)

(defmacro setv (l r)
 "like setq but nor result is returned.
  Useful at Top Level to supress too much output"
   `(progn (setq ,l ,r) nil)
)
         
;;;;-----------------------------------------------------------;;;;   
;;;;                   10. SYMBOLS                             ;;;;
;;;;-----------------------------------------------------------;;;;

(defmacro string-symbol (str) 
  "takes a string and returns the INTERNED SYMBOL whose 
  print name is STR (upcase)"
  `(intern (string-upcase ,str))
 )

(defmacro format-symbol (ctrl-string &rest args)
  "creates an interned symbol from a string generated by format"
   `(read-from-string (format nil ,ctrl-string ,@args))
 )

;;;;-----------------------------------------------------------;;;;   
;;;;                    15.  LISTS                             ;;;;
;;;;-----------------------------------------------------------;;;;

(defun flatten (lst)
 "returns the list of atoms which are recursively contained in lst"
   (if (not (listp lst))
       (list lst)
       (let (res)
            (dolist (l lst) (setq res (append res (flatten l))))
            res)
   ))
   
(defun permutate (lst)
 "returns a random permutation of the top level of the list lst
  top level is copied but no deeper levels, like append "
    (let ((i (length lst)))
         (do ((l nil (pushnew (nth (random i) lst) l)))
             ((eq (length l) i) l))))

(defun length< (lst1 lst2)
 "returns t if length of lst1 is smaller than length of lst2, else nil"
 ; especially useful as functional argument
   (< (length lst1)(length lst2))
  )
  
(defun length> (lst1 lst2)
 "returns t if length of lst1 is greater than length of lst2, else nil"
 ; especially useful as functional argument
   (> (length lst1)(length lst2))
  )
  
(defmacro pushl (l e)
 "pushs e as last element of l"
   (let ((l1 l))
      `(if ,l1 (push ,e (cdr (last ,l1))) (setf ,l1 (cons ,e nil)))
   )
)
  
;;;;-----------------------------------------------------------;;;;   
;;;;                      18. STRINGS                          ;;;;
;;;;-----------------------------------------------------------;;;;

(defmacro cat (&rest args)
  "concatenation of strings"
    `(concatenate 'string ,@args)
)

;;;;-----------------------------------------------------------;;;;   
;;;;                    22. INPUT/OUTPUT                       ;;;;
;;;;-----------------------------------------------------------;;;;


(defun format-table (destination list-of-lists 
                     &key (sep " | ") (mincol 0))
 "creates a table. Each element of list-of-lists corresponds to one line.
 If element is a list then each item of the list is separated by sep. 
 Seperators are only printed as long as tehere are items on the line.
 Thus an empty list prints only a newline. 
 If element is a character then the line is filled with that character.
 If mincol is a number then each column is at least mincol wide.
 If mincol is a list it gives the minimum width of the 
 corresponding columns and the last entry propagates to further columns."
  
 (let* ((max-length (apply #'max 
                          (mapcar #'(lambda (l)
                                       (if (listp l)(length l) 0))
                                  list-of-lists)))
        (sizes (if (numberp mincol)
                   (make-list max-length :initial-element mincol)
                   (append mincol 
                     (make-list (max (- max-length (length mincol)) 0)
                                :initial-element (car(last mincol))))))
        list-of-string-lists
        ctrl-string
       )
   ; make symbols to strings and set the maximal width for each column
  (setq list-of-string-lists 
    (mapcar
       #'(lambda (lst)
        (if (listp lst)
         (maplist
           #'(lambda (lst-rest sizes-rest)
            (let* ((entry (car lst-rest))
                   (str (if (stringp entry) 
                           entry  (write-to-string entry))) )
                (rplaca sizes-rest (max (length str)(car sizes-rest)))
                str))
           lst
           sizes)
          lst))
      list-of-lists))
   ; insert strings for the horizontal separator-lines
   ; (must be done after the width of columns has been determined)
  (setq list-of-string-lists 
    (mapcar
       #'(lambda (lst)
          (if (listp lst)
             lst
            (mapcar #'(lambda (i)(make-string i :initial-element lst ))
                    sizes)
         ))
        list-of-string-lists) )
   ; generate the control-string   
  (setq ctrl-string  ; "~{~{~<i1>A<sep>~^~<i2>A<sep>~^ ...~}~%~}" 
    (eval `(concatenate 'string 
              "~{~{"
              ,@(mapcar #'(lambda (i) (format nil "~~~DA~A~~^" i sep))
                        sizes) 
              "~}~%~}")) )
   ; print the table
  (if (bitmap-p destination) ; stringblt is much faster (3:5) than format
    (progn
      (stringblt destination (stream-position destination)
                 (stream-current-font destination) 
                 (format nil ctrl-string list-of-string-lists))
      (setf (windows::stream-position destination)
            (make-position 0 
                (+ (stream-y-position destination)
                   (* (length list-of-string-lists) 
                      (font-height (stream-current-font destination))))))
     )
    (format destination ctrl-string list-of-string-lists) )
 ))

(defun stop-until-enter ()
 "requests the user to press Enter and returns after having read a line"
  (format t "~%~%Press Enter to continue!")
  (read-line))
  
;;;  SAVING lisp OBJECTS

(defun save-SETQ (V &optional stream)  
 "saves the setting of the global variable V" 
  (print `(setq ,V ,(Kwote (symbol-value V)))
         stream)
  )
  ; remark: for structures with its last component being a list longer
  ;    than ca.30 chars, CL produces one additional right parantheses !?!
  ;    When read again or compiled this will produce warnings,
  ;    which could (and should) be ignored  

(defun save-DEFUN (F &optional stream)
 "saves the (interpreted!) function definition of F"
 (let ((fdef (symbol-function F))
       (doc_str (documentation F 'function)))
     (when (compiled-function-p fdef)
           (setq fdef (get F 'source-code)))
       ; fdef is an interpreted-function object
       ; extract now the defining function list
     (if fdef
         (progn
            (setq fdef ; make fdef a string and remove the header
                       ; prin1 must be used(!), format does not work for "".
                 (read-from-string (subseq (prin1-to-string fdef) 23)))
            (if doc_str
                (print `(defun ,F ,(third fdef) ,doc_str
                                  ,(third (fourth fdef))) 
                        stream)
                (print `(defun ,F ,(third fdef) 
                                  ,@(cddr (fourth fdef))) 
                       stream)))
         (format t "WARNING: ~A has no interpreted-function definition" F))
    ))
           
(defun save-DEFSTRUCT (S &optional stream)
 "saves the structure declaration for the structure S
  can NOT save defstruct-options nor slot-options (but saves defaults)!" 
 (let* ((accessor (read-from-string (format nil "make-~A" S)) )
         (slots (cdr (system::arglist accessor))) )
    (print `(defstruct ,S ,@slots)
           stream)
    ))

           
           
; -- to be continued --
             
;;;;-----------------------------------------------------------;;;;   
;;;;                23. FILE SYSTEM INTERFACE                  ;;;;
;;;;-----------------------------------------------------------;;;;

(defun fil-add-suffix (n s)
  "append `.s' to `n'"
   (concatenate 'string n "." s)
)

(defun fil-default-suffix (n s)
 "if `n' has a suffix, leave it as is, otherwise append `.s'"
 (if (pathname-type n)
     n
     (format nil "~A.~A" n s))
)

(defun fil-replace-suffix (n s)
  "if `n' has a suffix, replace it by `s', otherwise append `.s'"
   (let ((p (position #\. n)))
      (when p (setq n (subseq n 0 p)))
      (concatenate 'string n "." s)
   )
)

(defun fil-is-newer (file1 file2)
 "check whether both files exist, and file1 is newer than file2"
 (let ((f1 (full-pathname file1))  ; seems to be only necessary
        (f2 (full-pathname file2))  ; for CL V2.0 
                                    ; otherwise cd's are of no effect
                                    ; for file-write-date
       )
    (and (file-write-date f1)
	 (file-write-date f2)
	 (> (file-write-date f1) (file-write-date f2))
    )
  )
)

(defun path-conc (path &optional 
                       (prefix (namestring *default-pathname-defaults*)))
"concatenates the pathnames prefix (default for prefix is
 *default-pathname-defaults*) and path and returns that pathname"
   (pathname (concatenate 'string prefix path)))
   
   
(defun full-pathname (&optional (path nil))

"returns the FULL PATHNAME according to CD conventions:  
 concatenates the path to the current *default-pathname-defaults*, 
 unless path begins with / or ~ which means that the path is absolute 
 resp. relative to the home-directory.
 If there is no type supplied in path then path is took as directory and the
 returned string ends with /.
 If path is nil or empty then the homedir is returned"
 
 (let ((dir-list (pathname-directory path))
       name)
   ; first make the list of the correct dir-hierarchy   
  (cond ((or (null path)    ; path is relative to HOME-dir
             (equal path "")  
             (and (listp (car dir-list)) ; only CL V2.0
                  (equal (caar dir-list) ':home-dir))
            ) 
          (setq dir-list
            (append
              (pathname-directory (user-homedir-pathname))
              (cdr dir-list)))
            ; cdr because: in case 1 and 2 dir-list = (:relative)
            ;    and in case 3 dir-list = ((home-dir "") subdir-list)
          )
        ((eql (first dir-list) ':relative)
                   ; path is relative to CURRENT dir
          (setq dir-list
            (append
              (pathname-directory *default-pathname-defaults*)
              (cdr dir-list)))
          )
        (t  ) ; path is abolute
      )
  (when (eql (first dir-list) ':root) ; only  CL V2.1
     (setq dir-list (cdr dir-list))  ) 

  ; look for "." and ".."
  (setq dir-list (delete "." dir-list :test #'equal))  
  (loop
   (let ((pos (position ".." dir-list :test #'equal)))
     (unless pos (return))
     (setq dir-list 
         (nconc (subseq dir-list  0 (1- pos))
                (subseq dir-list (1+ pos)))  )
   ))
  ; now set the name if there is a type supplied
  (when path
     (if (pathname-type path)
       (setq name (format nil
                          "~A.~A" 
                          (pathname-name path)(pathname-type path)))
       (when (pathname-name path)
          (setq dir-list (nconc dir-list 
                                (list (pathname-name path)))))
   )  )
  ; return the full-pathname
  (format nil 
      "/~{~A/~}~@[~A~]"
       dir-list
       name   )
 ))   

;;;;-----------------------------------------------------------;;;;   
;;;;              25. MISCELLANEOUS FEATURES                   ;;;;
;;;;-----------------------------------------------------------;;;;

(defmacro msecs (sexpr &optional (tvar nil tvar_sup))
"evaluates sexpr and returns the result and as second value
 the cpu-time in msec needed for the evaluation.
 If tvar is supplied tvar is set to the cpu-time"
 `(let (old_time new_time used_time result)
      (setq old_time (get-internal-run-time))
      (setq result ,sexpr)
      (setq new_time (get-internal-run-time))
      (setq used_time 
         (round (/ (- new_time old_time)
                   (/ internal-time-units-per-second 1000))) )
      (when ,tvar_sup (setq ,tvar used_time))
     (values result used_time)
  ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;   
;;;;         import Lisp-extensions into package USER		;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'USER)
(use-package 'LISP-EXTENSION)

