;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;	  		 MODULE  M E N U    			;;;;
;;;;								;;;;
;;;;		(extension of Package WINDOWS)			;;;;
;;;;       (Chapter 12 of SUN COMMON LISP USER'S GUIDE)  	;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;           AUTHOR: Paul Weber, FZI, Karlsruhe  		;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (provide 'menu) at the end of file

(in-package 'windows)
(export '(
          make-menu display-menu clear-menu ; the main functions
          menu menu-p menu-width menu-height menu-x menu-y
          menu-cur-value menu-value menu-state menu-window
        ))




; ------------  INTRODUCTION  -------------
; A MENU-object is generated with MAKE-MENU.
; It can be displayed with DISPLAY-MENU
; and removed from screen with CLEAR-MENU.
; 
;     MENU - Handling 
;
; - Single-choice-menu
;    left-mouse-click in some field:
;      sets (overwrites) the value of the menu to the value of that field
;      and sets the state to t.
;      calls the ACTION associated with that value
;    middle-mouse-click in any field:
;      sets the value of the menu to NIL
;      and sets the state to t
;      calls the ACTION associated with NIL
; - Multiple-choice-menu
;    left-mouse-click in some field:
;      inverts the membership of the value of that field
;               in the value-list of the menu
;    middle-mouse-click in any field:
;      sets the value-list of the menu to NIL
;      calls the ACTION associated with NIL
;      and sets the state to t
;    right-mouse-click in any field:
;      and sets the state to t and thus confirms the selection
;      calls the ACTION associated with the menu
;
;  Remark: Multiple-choice-menus work only correct if all choice-values
;          are different!
;
;  On entering (exiting) some field cur-value of the menu is set to the 
;   value of that field (nil).
; 
; ----------- MENU-OBJECT ----------
; A MENU-OBJECT is a structure with slots:
;       window: window containing the menu for a choice-list
;       x,y   : the position where the menu was displayed last
;       width : width of the window
;       height: height of the window
;       multiple: indicates wether this is a single or multiple choice window
;       cur-value : is set to the value of the field the mouse is in
;       value : is set to the value(s) choosen from the menu by a
;               left-mouse-click(s) or to nil by a middle-mouse-click
;       state : is set to t by a right-mouse-click and also by a 
;               left-mouse-click in a single-choice menu
;
; -----------  MAKE-MENU  -----------
; PURPOSE: creates a Menu-object.
; SYNTAX : MAKE-MENU choice-list 
;                    &KEY x y width height font title border action
; REMARKS: choice-list is treated as by function make-pop-up-menu 
;          (cmp. Sun CL User's Guide pp. 12-77). Additionally each item 
;          could involve an action to be executed.
;          choise-item := atom | 
;                        ([string] atom [:action function/expression])
;  The window created is inactive, for activation see DISPLAY-MENU.
;  The window is as small as possible to map the complete
;  menu on its viewport.
;  x and y give the initial position for Display-menu.
;  Width and height are restrictions on the size of the resulting 
;   menu-window. If necessary the window has scroll-bars to fit
;   into this boundaries.
;   Make-menu returns NIL if the size of the window viewport
;   would be less than 5 in width or height!
;  Font is a font used for writing the choices and the title.
;  Title is a string which is treated as window title.
;  Border is a natural which gives the thickness of the window border.
; Defaults are: width = height = 2000,
;               x = y = 0
;               font = *default-font*, title = nil, border = 0.
;
; Actions: If no action is specified the mouse-clicks only affect the
;  state and value of the menu, which had to be tested and reacted to by
;  other program parts.
;  The specification of actions allows menus to act as an process
;  on their own.
;  After the keyword :action there must be an functional argument or
;  an expression.
;  Everytime the state of the menu is set to t, the action is executed.
;  If action is a function then this function is funcalled 
;  with the value of the menu as argument. If it is an expression this 
;  expression is just evaluated.
;  For single-valued menus different actions may be defined for 
;  the items within the choice-list by adding an action
;  in the defining list of the item.
;   
; Returns: the menu-object if successful, else nil.
;
; ----------  DISPLAY-MENU  -----------
; PURPOSE: displays the window of a Menu-object.
; SYNTAX : DISPLAY-MENU menu-object &optional x y 
; REMARKS: displays the window of menu with left-up-position (x,y),
;      if the window fits entirely on the root-vieport, else
;      nothing is displayed.
;      Perhaps the x(y)-coordinate is shifted to enable a full display of
;      the menu. Thus (display-menu menu 100000 100000) will display
;      the menu at the bottom-right-corner.
;      Calling DISPLAY-MENU again with no (other) coordinates will
;      just diplay the menu at the same (the other) position.
; Defaults for x (y) is the x(y)-slot of the menu
; Returns: the window, if it could be displayed, else nil.
;
; -----------  CLEAR-MENU  ------------
; PURPOSE: inactivates the window of a Menu-object.
; SYNTAX : CLEAR-MENU menu-object 
; REMARKS: inactivates the window of menu and sets the state
;          of the menu to nil. 
;          The menu can be displayed again with DISPLAY-MENU.
; Returns: the menu-value of menu.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct (MENU (:constructor create-menu))
   window
   (x 0) (y 0)  ; the last position of display
   width height
   (multiple nil)
   cur-value value state )

;; define type for items of the chooselist argument of make-menu
; choice-item := 
;    atom | ([string] atom [:action (function|expression)])
 (defun check-choice-item (l)
   (or (atom l)
       (and (consp l)
          (case (length l)
             (1 (and (atom (first l)) (null (cdr l))))
             (2 (string (first l)))
             (3 (and (atom (first l))
                     (eq (second l) ':action))) 
             (4 (and (string (first l))
                     (eq (third l) ':action)))
             (t nil)))
  ))
                     
 (deftype choice-item ()
     '(satisfies check-choice-item) )

(defun parse-choice-item (item)
 "returns name value and action of the item"
 (let (name value action)
   (when (and (listp item)(< (length item) 2))
         (setq item (car item)))
   (if (atom item)
      (values (format nil "~A" item) item nil)
      (let ((first (first item))
            (second (second item))
           ) 
        (if (stringp first)
          (setq name first)
          (setq name (format nil "~A" first)) )
        (if (eq second ':action)
          (setq value first)
          (setq value second) )
        (setq action (second (member ':action item)))
        (values name value action))
) ))
        
 
(defun MAKE-MENU (chooselist
                  &key (x 0) (y 0) ; initial display postition
                       (width  1000)
                       (height 1000); maximum width and height
                       (multiple nil)
                       (font *default-font*)
                       (title nil)
                       (border 1)
                       (action nil)
		       (with-static-area nil)
                       )
                       
  ;; type-checking of arguments
   (check-type chooselist list)
   (dolist (item chooselist)
     windows (check-type item choice-item))
   (check-type width (integer 0 *))
   (check-type height (integer 0 *))
   (check-type font font)
   (when title (check-type title string))
   (check-type border (integer 0 *))
   
  (prog*  ; enables premature return if sizes are too small for menu
       ((font-height (font-height font))
         (sep-width (round (* 0.1  font-height))) ;left/right-margin
         (line-height (+ font-height (* 2 sep-width)))
         (table-height 0)
         (table-width (if title (string-width title font) 0))
         (port-x border) ; port-offset within the window
         (port-y (+ border (if title font-height 0)))
         port-width port-height
         window-width window-height
         scroll-needed ; is set t if table is higher than height
         menu menu-window menu-bitmap
        )
      
  ;; determine width and height of a vertical table of items
   (dolist (item chooselist)
     (let ((name (parse-choice-item item)))
      (setq table-width 
           (max table-width (string-width name font)))
      (incf table-height line-height) ))
   (incf table-width (* 2 sep-width))
  ;; determine the port- and window-sizes 
   (setq window-width (+ table-width port-x border 2)) ; inner-border*2 =2
   (setq window-height (+ table-height port-y border 2))
   (when (> window-width width)
       (setq scroll-needed t)
       (setq window-width width))
   (when (> window-height height)
       (setq scroll-needed t)
       (setq window-height height))
   (setq port-width (- window-width port-x border
                                    (if scroll-needed 10 0) ))
   (setq port-height (- window-height port-y border
                                    (if scroll-needed 10 0) ))
  ;; return if port is too small
   (when (or (< port-height 5)(< port-width 5))
      (format t "Menu does not fit into a window of size ~D X ~D."
                window-width window-height)
      (return nil)) !!
            
  ;; create window for menu
   (setq menu-window
      (if with-static-area
	  (with-static-area
	      (make-window :viewport-x port-x
			   :viewport-y port-y
			   :viewport-width port-width
			   :viewport-height port-height
			   :width table-width :height table-height 
			   :outer-border-width border
			   :inner-border-width 1 ; must be >=1 in CL-2.1 (Bug?!)
			   :scroll scroll-needed
			   :title title
			   :title-font font
			   :activate nil ))
	   (make-window :viewport-x port-x
		       :viewport-y port-y
		       :viewport-width port-width
		       :viewport-height port-height
		       :width table-width :height table-height 
		       :outer-border-width border
		       :inner-border-width 1 ; must be >=1 in CL-2.1 (Bug?!)
		       :scroll scroll-needed
		       :title title
		       :title-font font
		       :activate nil )
       )
   )
   (setq menu-bitmap (viewport-bitmap menu-window))
   (setq menu (create-menu 
                     :window menu-window
                     :x x :y y
                     :width window-width
                     :multiple multiple
                     :height window-height))
       
    ;; define active regions for each item in the menu-bitmap
   (let ((count 0))
    (dolist (item chooselist)
     (let (name value val-action)
      (multiple-value-setq (name value val-action)
                              (parse-choice-item item))
      (unless val-action (setq val-action action))
      (if multiple
        ;; multiple-choice-active-regions
        (make-active-region
          (make-region :x 0 :y (* count line-height)
                       :width table-width :height line-height)
          :bitmap menu-bitmap
         :mouse-enter-region 
            #'(lambda (viewport active-region mouse-event x y)
               (invert-region viewport active-region mouse-event x y)
               (setf (menu-cur-value menu) value) )
         :mouse-exit-region 
            #'(lambda (viewport active-region mouse-event x y)
               (invert-region viewport active-region mouse-event x y)
               (setf (menu-cur-value menu) nil) )
         :mouse-left-down
            #'(lambda (viewport active-region mouse-event x y)
                 (declare (ignore mouse-event x y))
                (let ((menu-val (menu-value menu)))
                  (if (member value menu-val)
                    (setf (menu-value menu) (delete value menu-val))
                    (push  value (menu-value menu))) )
                (invert-region-partial viewport active-region
                    :left 0.02 :right 0.98 :bot 0.15 :top 0.85) )
          :mouse-middle-down
            #'(lambda (viewport active-region mouse-event x y)
                 (declare (ignore viewport active-region mouse-event x y))
                (setf (menu-value menu) nil)
                (setf (menu-state menu) t) 
                (if (and action (functionp action))
                    (funcall action nil)
                    ; (eval action)
                    ) )
          :mouse-right-down
            #'(lambda (viewport active-region mouse-event x y)
                 (declare (ignore viewport active-region mouse-event x y))
                (setf (menu-state menu) t)
                (if (and val-action (functionp val-action))
                    (funcall val-action (menu-value menu))
                    (eval val-action)) )
         )
        ;; single-choice-active-regions
        (make-active-region
          (make-region :x 0 :y (* count line-height)
                       :width table-width :height line-height)
          :bitmap menu-bitmap
          :mouse-enter-region 
            #'(lambda (viewport active-region mouse-event x y)
               (invert-region viewport active-region mouse-event x y)
               (setf (menu-cur-value menu) value) )
          :mouse-exit-region 
            #'(lambda (viewport active-region mouse-event x y)
               (invert-region viewport active-region mouse-event x y)
               (setf (menu-cur-value menu) nil) )
          :mouse-left-down
            #'(lambda (viewport active-region mouse-event x y)
                 (declare (ignore viewport active-region mouse-event x y))
                (setf (menu-value menu) value)
                (setf (menu-state menu) t) 
                (if (and val-action (functionp val-action))
                    (funcall val-action value)
                    (eval val-action)) )
          :mouse-middle-down
            #'(lambda (viewport active-region mouse-event x y)
                   (declare (ignore viewport active-region mouse-event x y))
                (setf (menu-value menu) nil)
                (setf (menu-state menu) t) 
                (if (and action (functionp action))
                    (funcall action nil)
                    ; (eval action)
                    ) )
         ))
      (stringblt menu-bitmap
        (make-position sep-width
                      (+ (* count line-height)
                         font-height) )
        font name)
      (incf count) )))
   (return menu)
 ))
 

(defun DISPLAY-MENU (menu &optional x y)
  ;; type-checking of arguments
   (check-type menu menu) 
   (when x (check-type x (integer 0 *)))
   (when y (check-type y (integer 0 *)))
  ;; body 
  (let* ((window (menu-window menu))
         (width  (menu-width menu))
         (height (menu-height menu))
         (xpos (if x x (menu-x menu)))
         (ypos (if y y (menu-y menu)))
         (root-bitmap (viewport-bitmap (root-viewport)))
         (rest-width (- (bitmap-width root-bitmap) xpos))
         (rest-height (- (bitmap-height root-bitmap) ypos))
        )
    (when (> width rest-width)
          (decf xpos (- width rest-width))) 
    (when (> height rest-height)
          (decf ypos (- height rest-height))) 
    (if (or (minusp xpos) (minusp ypos))
       (format t "Menu does not fit into root-viewport")
       (progn 
         (move-viewport window xpos ypos)
         (setf (menu-x menu) xpos)
         (setf (menu-y menu) ypos)
         (expose-viewport window)
         (activate-viewport window))
  )))
         
         
(defun CLEAR-MENU (menu)
   (check-type menu menu)  ;; type-checking of arguments 
  (deactivate-viewport (menu-window menu))
  (setf (menu-state menu) nil)
  (menu-value menu)
 )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;             E N D   M E N U                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'MENUES)
