;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;	   MODULE   W I N D O W - E X T E N S I O N		;;;;
;;;;								;;;;
;;;;								;;;;
;;;;       (Chapter 12 of SUN COMMON LISP USER'S GUIDE)  	;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;								;;;;
;;;;           AUTHOR: Paul Weber, FZI, Karlsruhe  		;;;;
;;;;								;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (provide 'window-extension) at the end of file

(in-package 'windows)
(export '(
          clear-root         
         ; Bitmap operations
          invert-region
          draw-rectangle
          stringblt-center string-into-region
          reset-bitmap         
         ; fonts ( in lib-module "load-fonts" )           
         ; mouse 
          define-cursors
          *std-cursor* *cross-cursor* *circles-cursor* 
                       *xor-cross-cursor* *xor-circles-cursor*

         ; Viewports
          move-viewport-center          
         ; Windows
          window-width window-height window-x window-y
          connect-windows disconnect-windows
          make-window-moveable
         ; Menues ( in lib-module "menues" )
	))
	
	

(defun CLEAR-ROOT ()
 ; deactivates all viewports on the root
 (dolist (v (viewport-children (root-viewport)))
    (deactivate-viewport v))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; functions for BITMAPs
;;;
                 
(defun RESET-BITMAP (bitmap)
 ; clears a bitmap and resets its stream-position to the origin
  (clear-bitmap bitmap)
  (setf (stream-position bitmap)
        (make-position 0 
                      (font-height (stream-current-font bitmap))
         )
   )
)

(defun INVERT-REGION (viewport region &optional mouse-event x y)
  ;; additional parameters for using it as a method for active regions 
     (declare (ignore mouse-event x y))
  (bitblt-region (viewport-bitmap viewport) region
                 (viewport-bitmap viewport) region
                 boole-c1)
  )

(defun invert-region-partial  ; not exported
    (viewport region
              &key (left 0) (right 1)
                   (bot 0)  (top 1) )
   (let ((reg-x (region-origin-x region))
         (reg-y (region-origin-y region))
         (reg-w (region-width region))
         (reg-h (region-height region)) )
     (let ((x (round (+ reg-x (* left reg-w))))
           (y (round (+ reg-y (* (- 1 top) reg-h))))
           (w (round (* reg-w (- right left))))
           (h (round (* reg-h (- top bot))))  )
        (bitblt viewport x y
                viewport x y
                w h
                boole-c1)
    )))    


(defun DRAW-RECTANGLE (bitmap origin corner 
                       &key (width 1) (operation boole-1))
 ; origin is the top-left-postion, corner the right-bottom-position
 ; for the other arguments compare pp. 12-51
   (let* ((left-bot-pos  (make-position (position-x origin)
                                        (position-y corner)))
          (right-top-pos (make-position (position-x corner)
                                        (position-y origin))) )
     (draw-polyline bitmap 
            (list origin right-top-pos corner
                         left-bot-pos origin)
            :width width :operation operation )
))

   
(defun STRINGBLT-CENTER (bitmap position font string 
                               &key (operation boole-1))
 "like stringblt, but centers string around position" 
 (let ((w (string-width string font))
       (h (font-baseline font)) )
    (stringblt bitmap
       (make-position (round (- (position-x position) (/ w 2)))
                      (round (+ (position-y position) (/ h 2))) )
       font string :operation operation)
  ))
  
  
(defun STRING-INTO-REGION (bitmap region fontlist string 
                               &key (operation boole-1))
 "Centers the string in the region. 
  Fontlist is a single font or a list of fonts.
  All given fonts are tried until the string fits into the region. When 
  none fits, the last font is taken and the string is positioned left.
  The fonts should be sorted in descending hugeness."
  
 (let ((w (region-width region))
       (h (region-height region))
      )
   (unless (listp fontlist) ; make a list of the single font
       (setq fontlist (list fontlist)))
   (unless 
     (some ; string fits into region for some font
       #'(lambda (font)
           (let ((sw (string-width string font))
                 (sh (font-baseline font)) 
                )
            (when (and (<= sw w) (<= sh h))
              (stringblt bitmap
                (make-position 
                   (round (+ (region-origin-x region) (/ (- w sw) 2)))
                   (round (+ (region-origin-y region) 
                             (/ (+ h sh) 2)
                             (- 1)) ) ; heuristic!
                 ) 
                font 
                string
                :operation operation)
              t
            )
           )
         )
       fontlist)
     ;; string does not string-into-region fit for any font. 
     ;; Thus position on left side when too long
     ;; resp. downside when too high
     (let* ((font (car (last fontlist)))
            (sw (string-width string font))
            (sh (font-baseline font))
            (b (make-bitmap :extent (region-size region))) 
           )
       (stringblt  b
         (make-position 
              (if (> sw w) 0 (round (/ (- w sw) 2)))
              (1- (if (> sh h) sh (round (/ (+ h sh) 2)))) )
          font 
          string
          :operation operation)
       (bitblt-region 
           b (make-region :x 0 :y 0 :extent (region-size region))
           bitmap region
           boole-ior)
     ) ; let*
   ) ; unless
  ) ; let
)
    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; functions for MOUSE-handling
;;;

(defvar *std-cursor*)
(defvar *cross-cursor*)
(defvar *circles-cursor*)
(defvar *xor-cross-cursor*)
(defvar *xor-circles-cursor*)

(defun define-cursors ()
 "defines some useful mouse-cursors.
  Call this function only after windows are initialized!!"
 
 ; define the standard cursor (assuming this is current)
  (setq *std-cursor* (current-mouse-cursor))
 ; define 2 cross-cursor
  (let ((mouse-bitmap (make-bitmap :width 16 :height 16)))      
    (draw-line mouse-bitmap (make-position 8 0)(make-position 8 16))
    (draw-line mouse-bitmap (make-position 0 8)(make-position 16 8))
    (setq *cross-cursor*
       (make-mouse-cursor mouse-bitmap :x-offset 8 :y-offset 8))
    (setq *xor-cross-cursor*
       (make-mouse-cursor mouse-bitmap :x-offset 8 :y-offset 8
                          :operation boole-xor))
  )
 ; define 2 circles-cursor
  (let ((mouse-bitmap (make-bitmap :width 16 :height 16)))      
    (draw-circle mouse-bitmap (make-position 8 8 ) 7 )
    (draw-circle mouse-bitmap (make-position 8 8 ) 5 )
    (draw-circle mouse-bitmap (make-position 8 8 ) 3 )
    (setq *circles-cursor*
      (make-mouse-cursor mouse-bitmap :x-offset 8 :y-offset 8))
    (setq *xor-circles-cursor*
      (make-mouse-cursor mouse-bitmap :x-offset 8 :y-offset 8
                         :operation boole-xor))
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; functions for VIEWPORTs
;;;

(defun MOVE-VIEWPORT-CENTER (viewport x y)
 "moves the viewport centered around (x,y) but x and y may be changed
  such that the viewport fits into the root-viewport.
  This is comparable to the positioning of pop-up-menus"                                    
 (let*  ((root (root-viewport))
         (rr (viewport-screen-region root))
         (rv (viewport-screen-region viewport))
         (w  (region-width rv))
         (h  (region-height rv))
         (max-x (- (region-width rr)  w))
         (max-y (- (region-height rr) h))
        )
   (setq x (- x (round (/ w 2))))
   (setq y (- y (round (/ h 2))))   
   (move-viewport viewport 
     (max
       (min x max-x) 
       0)
     (max
       (min y max-y)
       0)
   )
 )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; functions for WINDOWs
;;;

;; directly apply region-functions to windows

(defmacro window-width (w)
  `(region-width (viewport-screen-region ,w))
  )
(defmacro window-height(w)
  `(region-height (viewport-screen-region ,w))
  )
  
(defmacro window-x (w)
  `(region-origin-x (viewport-screen-region ,w))
  )
  
(defmacro window-y (w)
  `(region-origin-y (viewport-screen-region ,w))
  )
 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;;;Dis/Connecting windows
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
 
(defun CONNECT-WINDOWS (window-1 window-2 background-bitmap)
 ;; draw a line beetween two windows on the background bitmap.
   (draw-line background-bitmap
        (window-middle window-1) 
        (window-middle window-2))
 )

(defun DISCONNECT-WINDOWS (window-1 window-2 background-bitmap)
   (draw-line background-bitmap
        (window-middle window-1) 
        (window-middle window-2)
        :operation boole-c1      ) ; this also deletes cross-points of lines!
 )
 
 (defun window-middle (w)
  (let ((r (viewport-screen-region w)))
     (make-position
       (round
         (/ (+ (region-origin-x r)
               (region-corner-x r))
            2))         
       (round 
         (/ (+ (region-origin-y r)
               (region-corner-y r))
             2)))
  ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;; Making Windows/Viewports moveable with the mouse
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
   
(defun MAKE-WINDOW-MOVEABLE
      (window &key 
              region           ; the region sensitive for moving
                 ; Should be a region or the special keyword-value :frame.
                 ; Default region is the whole window
              (button :right)  ; the button to be pressed for moving
                 ; may be :left, :middle or :right, which is default.
              cursor           ; the cursor during moving
                 ; Default is the *xor-circles-cursor*.
              before           ; the function funcalled before moving
              after            ; the function funcalled after moving
                 ; these functions must have 5 arguments like
                 ; the methods for active-regions
  	 )
  "makes the window moveable while the mouse is pressed.
   returns the active region which makes the window moveable.
   Works also for viewports (then :frame is not allowed)."
  (let ((button-code
            (case button (:left 4)(:middle 2)(t 1)))
        (frame (eql region :frame))
        x-offset
        y-offset 
        (scrollable nil)     
       )
   (unless (boundp '*std-cursor*)
      (define-cursors)) ; Define-cursors if not done before
   (if (or frame (not (windowp window)))
     (setq x-offset 0 y-offset 0)
     (let ((o-b (window-outer-border-width window))
           (i-b (window-inner-border-width window))
           (t-b (when (window-title window)
                      (1+ (font-height (window-title-font window))) ))
           )
        (setq x-offset (+ i-b o-b) 
              y-offset (+ i-b (or t-b o-b) ))
        (when (window-vertical-scroll-ratio window)
           (setq scrollable t))
           ; in this case the bitmap-offset must be added dynamically
      )
    )
   (make-active-region
      (if frame 
        (make-region :extent (bitmap-extent (window-frame window)))
        (if region
           region            
           (make-region :extent (bitmap-extent window))  ) 
       )     
      :bitmap (if frame
                 (window-frame window)
                 window )
      (case button
         (:left   :mouse-left-down)
         (:middle :mouse-middle-down)
         (t      :mouse-right-down) )
       #'(lambda (w r m x y)
          (unwind-protect
            (progn
             (setf (current-mouse-cursor) 
                   (or cursor *xor-circles-cursor*))
             (when before (funcall before w r m x y))
             (loop
               (unless (eq (mouse-buttons) button-code) (return))
               (move-viewport window
                   (- (mouse-x)  x-offset
                      (- x (if scrollable (viewport-bitmap-x-offset w) 0)) )
                   (- (mouse-y) y-offset
                      (- y (if scrollable (viewport-bitmap-y-offset w) 0)) ) )
             )
            )
            (progn
             (when after (funcall after w r m x y))
             (setf (current-mouse-cursor) *std-cursor*)
            )
         ))
     ) ; make-active-region
 ); let
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  (provide 'window-extension)       ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

